import ibis
import pandas as pd
import pytest
from packaging import version

from xorq.common.utils.ibis_utils import from_ibis
from xorq.tests.util import assert_frame_equal, assert_series_equal


@pytest.mark.parametrize(
    "make_sort_key",
    (
        lambda t: ibis.asc("value"),
        lambda t: t.value.asc(),
        lambda t: ibis.desc("value"),
        lambda t: t.value.desc(),
        lambda t: ibis.asc(t.value, nulls_first=True),
    ),
)
def test_sort_keys(make_sort_key):
    data = pd.DataFrame({"id": [1, 2, 3, 4, 5], "value": [3.5, None, 4.8, 2.1, None]})
    t = ibis.memtable(data)
    expr = t.order_by(make_sort_key(t))

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


def test_cases():
    data = pd.DataFrame(
        {"int_col": [1, 2, 3, 1, 2], "string_col": ["a", "b", "c", "d", "e"]}
    )
    t = ibis.memtable(data)

    args = [(1, "abcd"), (2, "ABCD")]
    kwargs = {"else_": "dabc"}
    if version.parse(ibis.__version__) <= version.parse("9.5.0"):
        args = [[(1, "abcd"), (2, "ABCD")]]
        kwargs = {"default": "dabc"}

    expr = t.int_col.cases(*args, **kwargs)

    xorq_expr = from_ibis(expr)

    assert_series_equal(expr.execute(), xorq_expr.execute())


@pytest.mark.parametrize(
    "make_filter",
    (lambda t: t.int_col.isin((1, 2, 3)), lambda t: t.int_col.notin((1, 2, 3))),
)
def test_positional(make_filter):
    data = pd.DataFrame(
        {"int_col": [1, 2, 3, 4, 5], "string_col": ["a", "b", "c", "d", "e"]}
    )
    t = ibis.memtable(data)
    expr = t.filter(make_filter(t))

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


@pytest.mark.parametrize(
    "make_window",
    (
        lambda t: ibis.window(group_by=t.category, order_by=t.id),
        lambda t: ibis.window(
            group_by=t.category, order_by=t.id, preceding=2, following=0
        ),
        lambda t: None,
    ),
)
def test_window_with_rows(make_window):
    data = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6],
            "value": [10.0, 20.0, 30.0, 40.0, 50.0, 60.0],
            "category": ["A", "B", "A", "B", "A", "B"],
        }
    )
    t = ibis.memtable(data)

    expr = t.mutate(average=t.value.mean().over(make_window(t))).order_by("id")

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


def test_string_as_date():
    s = ibis.literal("2024-01-15")

    if version.parse(ibis.__version__) >= version.parse("11.0.0"):
        expr = s.as_date("%Y-%m-%d")
    else:
        expr = s.to_date("%Y-%m-%d")

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()

    if version.parse(ibis.__version__) >= version.parse("10.0.0"):
        actual = pd.to_datetime(actual)

    assert expected == actual


def test_string_as_timestamp():
    s = ibis.literal("2024-01-15 10:30:00")

    if version.parse(ibis.__version__) >= version.parse("11.0.0"):
        expr = s.as_timestamp("%Y-%m-%d %H:%M:%S")
    else:
        expr = s.to_timestamp("%Y-%m-%d %H:%M:%S")

    xorq_expr = from_ibis(expr)

    assert expr.execute() == xorq_expr.execute()


def test_integer_as_interval():
    i = ibis.literal(5)

    if version.parse(ibis.__version__) >= version.parse("10.0.0"):
        expr = i.as_interval("day")
    else:
        expr = i.to_interval(unit="day")

    xorq_expr = from_ibis(expr)

    assert expr.execute() == xorq_expr.execute()


def test_integer_as_timestamp():
    i = ibis.literal(1705315800)

    if version.parse(ibis.__version__) >= version.parse("10.0.0"):
        expr = i.as_timestamp("s")
    else:
        expr = i.to_timestamp(unit="s")

    xorq_expr = from_ibis(expr)

    assert expr.execute() == xorq_expr.execute()


def test_table_unpack():
    data = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "data": [{"x": 10, "y": "a"}, {"x": 20, "y": "b"}, {"x": 30, "y": "c"}],
        }
    )
    t = ibis.memtable(data)

    if version.parse(ibis.__version__) >= version.parse("11.0.0"):
        expr = t.unpack("data")
    else:
        expr = t.select(t.id, *t.data.destructure())

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


def test_limit_positional():
    data = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            "value": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0],
        }
    )
    t = ibis.memtable(data)
    expr = t.limit(10)

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


def test_limit_offset_keyword():
    data = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            "value": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0],
        }
    )
    t = ibis.memtable(data)
    expr = t.limit(10, offset=5)

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


def test_identical_to_positional():
    data = pd.DataFrame({"id": [1, 2, None, 4, 1], "value": [1.0, 2.0, 3.0, 4.0, 5.0]})
    t = ibis.memtable(data)
    expr = t.filter(t.id.identical_to(1))

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())



import pytest

from xorq.common.utils.ibis_utils import from_ibis


ibis = pytest.importorskip("ibis")


def test_add():
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 + lit2

    xo_expr = from_ibis(expr)
    assert xo_expr is not None


def test_subtract():
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 - lit2

    xo_expr = from_ibis(expr)
    assert xo_expr is not None


def test_multiply():
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 * lit2

    xo_expr = from_ibis(expr)
    assert xo_expr is not None


def test_divide():
    lit1 = ibis.literal(6.0)
    lit2 = ibis.literal(2.0)
    expr = lit1 / lit2

    xo_expr = from_ibis(expr)
    assert xo_expr is not None


def test_mixed_arithmetic():
    i = ibis.literal(5)
    f = ibis.literal(2.5)
    expr = i * f

    xo_expr = from_ibis(expr)
    assert xo_expr is not None


def test_complex_arithmetic():
    a = ibis.literal(10)
    b = ibis.literal(5)
    c = ibis.literal(2.0)
    expr = (a + b) * c

    xo_expr = from_ibis(expr)
    assert xo_expr is not None



import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
from xorq.tests.util import assert_series_equal


def test_map_get_with_compatible_value_smaller(con):
    value = xo.literal({"A": 1000, "B": 2000})
    expr = value.get("C", 3)
    assert con.execute(expr) == 3


def test_map_get_with_compatible_value_bigger(con):
    value = xo.literal({"A": 1, "B": 2})
    expr = value.get("C", 3000)
    assert con.execute(expr) == 3000


def test_map_get_with_incompatible_value_different_kind(con):
    value = xo.literal({"A": 1000, "B": 2000})
    expr = value.get("C", 3.0)
    assert con.execute(expr) == 3.0


@pytest.mark.parametrize(
    ("map", "key"),
    [
        param(
            xo.map(
                xo.literal(["a", "b"]), xo.literal(["c", "d"], type="array<string>")
            ),
            xo.literal(None, type="string"),
            id="non_null_map_null_key",
        ),
        param(
            xo.map(
                xo.literal(None, type="array<string>"),
                xo.literal(None, type="array<string>"),
            ),
            xo.literal(None, type="string"),
            id="null_both_null_key",
        ),
    ],
)
@pytest.mark.parametrize("method", ["get", "contains"])
def test_map_get_contains_nulls(con, map, key, method):
    expr = getattr(map, method)
    assert con.execute(expr(key)) is None


def test_scalar_isin_literal_map_keys(con):
    mapping = xo.literal({"a": 1, "b": 2})
    a = xo.literal("a")
    c = xo.literal("c")
    true = a.isin(mapping.keys())
    false = c.isin(mapping.keys())
    assert con.execute(true) == True  # noqa: E712
    assert con.execute(false) == False  # noqa: E712


def test_map_scalar_contains_key_scalar(con):
    mapping = xo.literal({"a": 1, "b": 2})
    a = xo.literal("a")
    c = xo.literal("c")
    true = mapping.contains(a)
    false = mapping.contains(c)
    assert con.execute(true) == True  # noqa: E712
    assert con.execute(false) == False  # noqa: E712


def test_map_scalar_contains_key_column(alltypes, alltypes_df):
    value = {"1": "a", "3": "c"}
    mapping = xo.literal(value)
    expr = mapping.contains(alltypes.string_col).name("tmp")
    result = expr.execute()
    expected = alltypes_df.string_col.apply(lambda x: x in value).rename("tmp")
    assert_series_equal(result, expected)


def test_map_column_contains_key_scalar(alltypes, alltypes_df):
    expr = xo.map(xo.array([alltypes.string_col]), xo.array([alltypes.int_col]))
    series = alltypes_df.apply(lambda row: {row["string_col"]: row["int_col"]}, axis=1)

    result = expr.contains("1").name("tmp").execute()
    series = series.apply(lambda x: "1" in x).rename("tmp")

    assert_series_equal(result, series)


def test_map_column_contains_key_column(alltypes):
    map_expr = xo.map(xo.array([alltypes.string_col]), xo.array([alltypes.int_col]))
    expr = map_expr.contains(alltypes.string_col).name("tmp")
    result = expr.execute()
    assert result.all()


@pytest.mark.parametrize(
    ("keys", "values"),
    [
        param(["a", "b"], [1, 2], id="string"),
        param(["a", "b"], ["1", "2"], id="int"),
    ],
)
def test_map_construct_dict(con, keys, values):
    expr = xo.map(keys, values)
    result = con.execute(expr.name("tmp"))
    assert result == dict(zip(keys, values))


def test_map_construct_array_column(con, alltypes, alltypes_df):
    expr = xo.map(xo.array([alltypes.string_col]), xo.array([alltypes.int_col]))
    result = con.execute(expr)
    expected = alltypes_df.apply(
        lambda row: {row["string_col"]: row["int_col"]}, axis=1
    )

    assert result.to_list() == expected.to_list()


@pytest.mark.parametrize("null_value", [None, xo.null()], ids=["none", "null"])
def test_map_get_with_null_on_null_type_with_null(con, null_value):
    value = xo.literal({"A": None, "B": None})
    expr = value.get("C", null_value)
    result = con.execute(expr)
    assert pd.isna(result)


def test_map_contains_null(con):
    expr = xo.map(["a"], xo.literal([None], type="array<string>"))
    assert con.execute(expr.contains("a"))
    assert not con.execute(expr.contains("b"))


def test_map_length(con):
    expr = xo.literal({"a": "A", "b": "B"}).length()
    assert con.execute(expr) == 2



from typing import Any, OrderedDict, Tuple

import xorq.expr.relations as rel
import xorq.expr.udf as udf
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis import Expr
from xorq.vendor.ibis.expr.operations.core import Node


opaque_ops = (
    rel.Read,
    rel.CachedNode,
    rel.RemoteTable,
    rel.FlightUDXF,
    rel.FlightExpr,
    udf.ExprScalarUDF,
)


def to_node(maybe_expr: Any) -> Node:
    match maybe_expr:
        case Node():
            return maybe_expr
        case Expr():
            return maybe_expr.op()
        case _:
            raise ValueError(f"Don't know how to handle type {type(maybe_expr)}")


def gen_children_of(node: Node) -> Tuple[Node, ...]:
    match node:
        case ops.Field():
            rel_node = node.rel
            gen = () if rel_node is None else (to_node(rel_node),)

        case rel.RemoteTable():
            gen = (to_node(node.remote_expr),)

        case rel.CachedNode():
            gen = (to_node(node.parent),)

        case rel.FlightExpr() | rel.FlightUDXF():
            gen = (to_node(node.input_expr),)

        case udf.ExprScalarUDF():
            gen = (to_node(node.computed_kwargs_expr),)

        case rel.Read():
            gen = ()
        case _:
            raw_children = getattr(node, "__children__", ())
            gen = map(to_node, raw_children)
    yield from filter(None, gen)


def bfs(node):
    from collections import deque

    from xorq.vendor.ibis.common.graph import Graph

    queue = deque((to_node(node),))
    dct = {}
    while queue:
        if (node := queue.popleft()) not in dct:
            children = tuple(gen_children_of(node))
            dct[node] = children
            queue.extend(children)
    return Graph(dct)


def walk_nodes(node_types, expr):
    # TODO should this function use an ordered set
    visited = set()
    to_visit = [to_node(expr)]
    result = ()

    while to_visit:
        node = to_visit.pop()
        if node in visited:
            continue
        visited.add(node)
        if isinstance(node, node_types):
            result += (node,)

        to_visit += (
            child
            for child in OrderedDict.fromkeys(gen_children_of(node))
            if child not in visited
        )

    return result


def replace_nodes(replacer, expr):
    # Cache results of opaque sub-expression traversals by their root node.
    # Sub-expression roots are often shared across multiple opaque nodes (e.g.
    # each pipeline step's ExprScalarUDF references accumulated sub-expressions
    # that overlap heavily), so without this memo each shared root gets
    # re-traversed once per reference — O(n²) for a depth-n pipeline.
    sub_expr_memo = {}

    def do_recreate(op, _kwargs, **kwargs):
        kwargs = dict(zip(op.__argnames__, op.__args__)) | (_kwargs or {}) | kwargs
        return op.__recreate__(kwargs)

    def _replace_sub(sub_op):
        if sub_op not in sub_expr_memo:
            sub_expr_memo[sub_op] = sub_op.replace(process_node).to_expr()
        return sub_expr_memo[sub_op]

    def process_node(op, _kwargs):
        op = replacer(op, _kwargs)
        match op:
            case rel.RemoteTable():
                remote_expr = _replace_sub(op.remote_expr.op())
                return do_recreate(op, _kwargs, remote_expr=remote_expr)
            case rel.CachedNode():
                parent = _replace_sub(op.parent.op())
                return do_recreate(op, _kwargs, parent=parent)
            case rel.FlightExpr() | rel.FlightUDXF():
                input_expr = _replace_sub(op.input_expr.op())
                return do_recreate(op, _kwargs, input_expr=input_expr)
            case udf.ExprScalarUDF():
                computed_kwargs_expr = _replace_sub(op.computed_kwargs_expr.op())
                with_cke = op.with_computed_kwargs_expr(computed_kwargs_expr)
                return do_recreate(with_cke, _kwargs)
            case rel.Read():
                return op
            case _:
                if isinstance(op, opaque_ops):
                    raise ValueError(f"unhandled opaque op {type(op)}")
                return op

    initial_op = expr.op() if hasattr(expr, "op") else expr
    op = initial_op.replace(process_node)
    return op


def get_ordered_unique_sources(nodes):
    sources, seen = (), set()
    for source in (node.source for node in nodes):
        if source not in seen:
            seen.add(source)
            sources += (source,)
    return sources


def find_all_sources(expr):
    import xorq.vendor.ibis.expr.operations as ops

    node_types = (
        ops.DatabaseTable,
        ops.SQLQueryResult,
        rel.CachedNode,
        rel.Read,
        rel.RemoteTable,
        rel.FlightUDXF,
        rel.FlightExpr,
        # ExprScalarUDF has an expr we need to get to
        # FlightOperator has a dynamically generated connection: it should be passed a Profile instead
    )
    nodes = walk_nodes(node_types, expr)
    sources = get_ordered_unique_sources(nodes)
    return sources



from toolz import compose


convert_sorted_kwargs_tuple = compose(tuple, sorted, dict.items, dict)


def validate_kwargs_tuple(instance, attribute, value):
    assert isinstance(value, tuple) and all(
        isinstance(el, tuple) and len(el) == 2 for el in value
    )



from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import pandas as pd
import pyarrow as pa
from pyiceberg.catalog import load_catalog
from pyiceberg.table import ALWAYS_TRUE
from pyiceberg.table import Table as IcebergTable

import xorq.vendor.ibis.expr.operations as ops
from xorq.backends.postgres.compiler import compiler as postgres_compiler
from xorq.backends.pyiceberg.compiler import PyIceberg, translate
from xorq.backends.pyiceberg.relations import PyIcebergTable
from xorq.vendor import ibis
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.expr import schema as sch
from xorq.vendor.ibis.expr import types as ir
from xorq.vendor.ibis.util import gen_name


def parse_url(url: str) -> Dict[str, Any]:
    from urllib.parse import parse_qs, urlparse

    parsed = urlparse(url)
    warehouse_path = (
        parsed.netloc + parsed.path if parsed.netloc else parsed.path.lstrip("/")
    )
    params = {k: v[0] if len(v) == 1 else v for k, v in parse_qs(parsed.query).items()}

    return {
        **{
            "namespace": "default",
            "catalog_name": "default",
            "catalog_type": "sql",
            "warehouse_path": warehouse_path,
        },
        **params,
    }


def _overwrite_table_data(iceberg_table: IcebergTable, data: pa.Table):
    tx = iceberg_table.transaction()
    tx.delete(delete_filter=ALWAYS_TRUE)

    update_snapshot = tx.update_snapshot()
    with update_snapshot.fast_append() as append_files:
        for data_file in iceberg_table.writer().write_table(data):
            append_files.append_data_file(data_file)

    tx.commit_transaction()


class Backend(SQLBackend):
    name = "pyiceberg"
    dialect = PyIceberg
    compiler = postgres_compiler

    @property
    def version(self):
        import importlib.metadata

        return importlib.metadata.version("pyiceberg")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.catalog = None
        self.namespace = "default"
        self.warehouse_path = None
        self.duckdb_con = None
        self.catalog_params = None
        self.uri = None

    @staticmethod
    def _from_url(url: str) -> Dict[str, Any]:
        return parse_url(url)

    @classmethod
    def connect_env(cls, **kwargs):
        from xorq.common.utils.pyiceberg_utils import make_connection

        return make_connection(**kwargs)

    def do_connect(
        self,
        uri=None,
        warehouse_path="warehouse",
        namespace="default",
        catalog_name="default",
        catalog_type="sql",
    ) -> None:
        self.warehouse_path = Path(warehouse_path).absolute()
        self.namespace = namespace
        self.warehouse_path.mkdir(exist_ok=True)
        self.uri = uri or f"sqlite:///{self.warehouse_path}/pyiceberg_catalog.db"

        self.catalog_params = {
            "type": catalog_type,
            "uri": self.uri,
            "warehouse": f"file://{self.warehouse_path}",
        }

        self.catalog = load_catalog(catalog_name, **self.catalog_params)

        if self.namespace not in [n[0] for n in self.catalog.list_namespaces()]:
            self.catalog.create_namespace(self.namespace)

    def table(
        self,
        name: str,
        schema: str | None = None,
        database: tuple[str, str] | str | None = None,
        snapshot_id: str | None = None,
    ) -> ir.Table:
        table_schema = self.get_schema(name, catalog=schema, database=database)

        return PyIcebergTable(
            name=name,
            schema=table_schema,
            source=self,
            namespace=ops.Namespace(catalog=schema, database=database),
            snapshot_id=snapshot_id,
        ).to_expr()

    def create_table(
        self,
        name: str,
        obj: Optional[Union[pd.DataFrame, pa.Table, ir.Table]] = None,
        *,
        schema: Optional[ibis.Schema] = None,
        database: Optional[str] = None,
        temp: bool = False,
        overwrite: bool = False,
    ) -> ir.Table:
        database = database or self.namespace
        full_table_name = f"{database}.{name}"

        if self.catalog.table_exists(full_table_name):
            if not overwrite:
                raise ValueError(f"Table {full_table_name} already exists")
            self.drop_table(name, database=database)

        if obj is None and schema is not None:
            raise NotImplementedError("Creating empty tables is not implemented yet")

        if isinstance(obj, pd.DataFrame):
            obj = pa.Table.from_pandas(obj)
        elif isinstance(obj, ir.Table):
            obj = self.to_pyarrow(obj)

        # NEEDS this to reset the field_id
        obj = obj.cast(
            pa.schema(
                [
                    pa.field(name, type=typ)
                    for name, typ in zip(obj.schema.names, obj.schema.types)
                ]
            )
        )

        iceberg_table = self.catalog.create_table(
            identifier=full_table_name, schema=obj.schema
        )
        iceberg_table.append(obj)

        return self.table(name)

    def insert(
        self,
        table_name: str,
        data: Union[pd.DataFrame, pa.Table, ir.Table],
        database: Optional[str] = None,
        mode: str = "append",
    ) -> bool:
        database = database or self.namespace
        full_table_name = f"{database}.{table_name}"

        if not self.catalog.table_exists(full_table_name):
            raise ValueError(f"Table {full_table_name} does not exist")

        if isinstance(data, pd.DataFrame):
            data = pa.Table.from_pandas(data)
        elif isinstance(data, ir.Table):
            data = self.to_pyarrow(data)

        iceberg_table = self.catalog.load_table(full_table_name)
        iceberg_table.refresh()

        if mode == "overwrite":
            _overwrite_table_data(iceberg_table, data)
        else:
            with iceberg_table.transaction() as transaction:
                transaction.append(data)

        return True

    def upsert(
        self,
        table_name: str,
        data: Union[pd.DataFrame, pa.Table, ir.Table],
        database: Optional[str] = None,
        join_cols=None,
        when_matched_update_all=True,
        when_not_matched_insert_all=True,
        case_sensitive=True,
    ):
        """Wrapper around upsert"""
        database = database or self.namespace
        full_table_name = f"{database}.{table_name}"

        if isinstance(data, pd.DataFrame):
            data = pa.Table.from_pandas(data)
        elif isinstance(data, ir.Table):
            data = self.to_pyarrow(data)

        iceberg_table = self.catalog.load_table(full_table_name)
        iceberg_table.upsert(
            data,
            join_cols=join_cols,
            when_matched_update_all=when_matched_update_all,
            when_not_matched_insert_all=when_not_matched_insert_all,
            case_sensitive=case_sensitive,
        )

        return self.table(table_name)

    def to_pyarrow(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        batches = self.to_pyarrow_batches(expr, params=params, limit=limit, **kwargs)
        return batches.read_all()

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Optional[Mapping[ir.Scalar, Any]] = None,
        limit: Optional[Union[int, str]] = None,
        chunk_size: int = 10_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        query = translate(expr.op(), namespace=self.namespace, catalog=self.catalog)
        return query.to_arrow_batch_reader()

    def list_tables(
        self,
        like: Optional[str] = None,
        database: Optional[Union[Tuple[str, str], str]] = None,
    ) -> List[str]:
        database = database or self.namespace
        table_names = [t[1] for t in self.catalog.list_tables(database)]

        if like is not None:
            import fnmatch

            return [t for t in table_names if fnmatch.fnmatch(t, like)]

        return table_names

    def drop_table(
        self,
        name: str,
        database: Optional[str] = None,
        force: bool = False,
    ) -> None:
        database = database or self.namespace
        full_table_name = f"{database}.{name}"

        if self.catalog.table_exists(full_table_name):
            self.catalog.drop_table(full_table_name)

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: Optional[str] = None,
        database: Optional[str] = None,
    ) -> sch.Schema:
        database = database or self.namespace
        catalog = (
            self.catalog
            if catalog is None
            else load_catalog(catalog, **self.catalog_params)
        )

        table = catalog.load_table(f"{database}.{table_name}")
        return sch.Schema.from_pyarrow(table.schema().as_arrow())

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        raise NotImplementedError("_get_schema_using_query")

    def read_record_batches(
        self,
        reader: Union[pa.RecordBatchReader, pa.ChunkedArray],
        table_name: Optional[str] = None,
    ) -> ir.Table:
        table_name = table_name or gen_name("read_record_batches")
        table = pa.Table.from_batches(reader, reader.schema)

        self.create_table(name=table_name, obj=table, database=self.namespace)
        return self.table(table_name)

    def list_snapshots(self, database=None) -> dict[str, int]:
        database = database or self.namespace
        table_names = [t[1] for t in self.catalog.list_tables(database)]

        snapshots = {}
        for table_name in table_names:
            ice_table = self.catalog.load_table(f"{database}.{table_name}")
            ice_table.inspect.snapshots()
            snapshots[table_name] = (
                ice_table.inspect.snapshots().column("snapshot_id").to_pylist()
            )

        return snapshots



import pyarrow as pa

import xorq.api as xo
from xorq.backends.pyiceberg.tests.conftest import QUOTES_TABLE_NAME
from xorq.caching import SourceCache
from xorq.tests.util import assert_frame_equal


def test_into_backend(iceberg_con, trades_df):
    con = xo.connect()
    t = con.create_table("trades", trades_df)
    expr = t.select(t.symbol, t.price, t.volume).into_backend(
        iceberg_con, "xorq_trades"
    )

    expected = trades_df[["symbol", "price", "volume"]]

    actual = expr.execute()

    assert_frame_equal(actual, expected)


def test_out_into_backend(quotes_table, quotes_df):
    con = xo.connect()
    expr = quotes_table.select("symbol", "bid").into_backend(con, name="iceberg_quotes")
    expected = quotes_df[["symbol", "bid"]]
    actual = expr.execute()

    assert_frame_equal(actual, expected)


def test_caching(iceberg_con, quotes_table):
    cache = SourceCache.from_kwargs(source=iceberg_con)
    uncached_expr = quotes_table.select("symbol", "bid").filter(xo._.symbol == "GOOGL")
    expr = uncached_expr.cache(cache)

    assert not cache.exists(uncached_expr)
    assert not expr.execute().empty
    assert cache.exists(uncached_expr)


def test_upsert(iceberg_con, quotes_table):
    arrow_schema = pa.schema(
        [
            pa.field("city", pa.string(), nullable=False),
            pa.field("inhabitants", pa.int32(), nullable=False),
        ]
    )

    df = pa.Table.from_pylist(
        [
            {"city": "Amsterdam", "inhabitants": 921402},
            {"city": "San Francisco", "inhabitants": 808988},
            {"city": "Drachten", "inhabitants": 45019},
            {"city": "Paris", "inhabitants": 2103000},
        ],
        schema=arrow_schema,
    )

    table = iceberg_con.create_table("cities", df)
    before = table.execute()

    df = pa.Table.from_pylist(
        [
            {"city": "Drachten", "inhabitants": 45505},
            {"city": "Berlin", "inhabitants": 3432000},
            {"city": "Paris", "inhabitants": 2103000},
        ],
        schema=arrow_schema,
    )

    table = iceberg_con.upsert("cities", df, join_cols=["city"])
    after = table.execute()

    assert not before.equals(after)
    actual = after.set_index("city").to_dict()["inhabitants"]

    assert "Berlin" in actual
    assert actual["Drachten"] == 45505


def test_list_snapshots(iceberg_con):
    actual = iceberg_con.list_snapshots()
    assert isinstance(actual, dict)


def test_filter_by_snapshot_simple(iceberg_con):
    lookup = iceberg_con.list_snapshots()
    first, *_ = lookup[QUOTES_TABLE_NAME]

    table = iceberg_con.table(QUOTES_TABLE_NAME, snapshot_id=first)
    assert not table.execute().empty


def test_filter_by_snapshot(iceberg_con, quotes_df):
    table_name = "double_quotes"
    iceberg_con.create_table(table_name, quotes_df)
    quotes_double_bid = quotes_df.assign(bid=quotes_df.bid * 2)
    iceberg_con.insert(table_name, quotes_double_bid)
    lookup = iceberg_con.list_snapshots()
    first_id, second_id = lookup[table_name]

    first = iceberg_con.table(table_name, snapshot_id=first_id).execute()
    actual = iceberg_con.table(table_name, snapshot_id=second_id).execute()

    assert not first["bid"].equals(actual["bid"])



import inspect
from functools import (
    partial,
)
from importlib import (
    import_module,
)

from cloudpickle import (
    dumps,
    loads,
)
from toolz.functoolz import (
    _sigs,
    has_varargs,
    instanceproperty,
    is_partial_args,
    is_valid_args,
    no_default,
)


class curry(object):
    """Curry a callable function

    Enables partial application of arguments through calling a function with an
    incomplete set of arguments.

    >>> def mul(x, y):
    ...     return x * y
    >>> mul = curry(mul)

    >>> double = mul(2)
    >>> double(10)
    20

    Also supports keyword arguments

    >>> @curry                  # Can use curry as a decorator
    ... def f(x, y, a=10):
    ...     return a * (x + y)

    >>> add = f(a=1)
    >>> add(2, 3)
    5

    See Also:
        toolz.curried - namespace of curried functions
                        https://toolz.readthedocs.io/en/latest/curry.html
    """

    def __init__(self, *args, **kwargs):
        if not args:
            raise TypeError("__init__() takes at least 2 arguments (1 given)")
        func, args = args[0], args[1:]
        if not callable(func):
            raise TypeError("Input must be callable")

        # curry- or functools.partial-like object?  Unpack and merge arguments
        if (
            hasattr(func, "func")
            and hasattr(func, "args")
            and hasattr(func, "keywords")
            and isinstance(func.args, tuple)
        ):
            _kwargs = {}
            if func.keywords:
                _kwargs.update(func.keywords)
            _kwargs.update(kwargs)
            kwargs = _kwargs
            args = func.args + args
            func = func.func

        if kwargs:
            self._partial = partial(func, *args, **kwargs)
        else:
            self._partial = partial(func, *args)

        self.__doc__ = getattr(func, "__doc__", None)
        self.__name__ = getattr(func, "__name__", "<curry>")
        self.__module__ = getattr(func, "__module__", None)
        self.__qualname__ = getattr(func, "__qualname__", None)
        self._sigspec = None
        self._has_unknown_args = None

    @instanceproperty
    def func(self):
        return self._partial.func

    @instanceproperty
    def __signature__(self):
        sig = inspect.signature(self.func)
        args = self.args or ()
        keywords = self.keywords or {}
        if is_partial_args(self.func, args, keywords, sig) is False:
            raise TypeError("curry object has incorrect arguments")

        params = list(sig.parameters.values())
        skip = 0
        for param in params[: len(args)]:
            if param.kind == param.VAR_POSITIONAL:
                break
            skip += 1

        kwonly = False
        newparams = []
        for param in params[skip:]:
            kind = param.kind
            default = param.default
            if kind == param.VAR_KEYWORD:
                pass
            elif kind == param.VAR_POSITIONAL:
                if kwonly:
                    continue
            elif param.name in keywords:
                default = keywords[param.name]
                kind = param.KEYWORD_ONLY
                kwonly = True
            else:
                if kwonly:
                    kind = param.KEYWORD_ONLY
                if default is param.empty:
                    default = no_default
            newparams.append(param.replace(default=default, kind=kind))

        return sig.replace(parameters=newparams)

    @instanceproperty
    def args(self):
        return self._partial.args

    @instanceproperty
    def keywords(self):
        return self._partial.keywords

    @instanceproperty
    def func_name(self):
        return self.__name__

    def __str__(self):
        return str(self.func)

    def __repr__(self):
        return repr(self.func)

    def __hash__(self):
        return hash(
            (
                self.func,
                self.args,
                frozenset(self.keywords.items()) if self.keywords else None,
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, curry)
            and self.func == other.func
            and self.args == other.args
            and self.keywords == other.keywords
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __call__(self, *args, **kwargs):
        try:
            return self._partial(*args, **kwargs)
        except TypeError as exc:
            if self._should_curry(args, kwargs, exc):
                return self.bind(*args, **kwargs)
            raise

    def _should_curry(self, args, kwargs, exc=None):
        func = self.func
        args = self.args + args
        if self.keywords:
            kwargs = dict(self.keywords, **kwargs)
        if self._sigspec is None:
            sigspec = self._sigspec = _sigs.signature_or_spec(func)
            self._has_unknown_args = has_varargs(func, sigspec) is not False
        else:
            sigspec = self._sigspec

        if is_partial_args(func, args, kwargs, sigspec) is False:
            # Nothing can make the call valid
            return False
        elif self._has_unknown_args:
            # The call may be valid and raised a TypeError, but we curry
            # anyway because the function may have `*args`.  This is useful
            # for decorators with signature `func(*args, **kwargs)`.
            return True
        elif not is_valid_args(func, args, kwargs, sigspec):
            # Adding more arguments may make the call valid
            return True
        else:
            # There was a genuine TypeError
            return False

    def bind(self, *args, **kwargs):
        return type(self)(self, *args, **kwargs)

    def call(self, *args, **kwargs):
        return self._partial(*args, **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return curry(self, instance)

    def __reduce__(self):
        func = self.func
        modname = getattr(func, "__module__", None)
        qualname = getattr(func, "__qualname__", None)
        if qualname is None:  # pragma: no cover
            qualname = getattr(func, "__name__", None)
        is_decorated = None
        if modname and qualname:
            if modname == "__main__":
                func = dumps(func)
            else:
                attrs = []
                obj = import_module(modname)
                for attr in qualname.split("."):
                    if isinstance(obj, curry):
                        attrs.append("func")
                        obj = obj.func
                    obj = getattr(obj, attr, None)
                    if obj is None:
                        break
                    attrs.append(attr)
                if isinstance(obj, curry) and obj.func is func:
                    is_decorated = obj is self
                    qualname = ".".join(attrs)
                    func = "%s:%s" % (modname, qualname)

        # functools.partial objects can't be pickled
        userdict = tuple(
            (k, v)
            for k, v in self.__dict__.items()
            if k not in ("_partial", "_sigspec")
        )
        state = (type(self), func, self.args, self.keywords, userdict, is_decorated)
        return _restore_curry, state


def _restore_curry(cls, func, args, kwargs, userdict, is_decorated):
    if isinstance(func, str):
        modname, qualname = func.rsplit(":", 1)
        obj = import_module(modname)
        for attr in qualname.split("."):
            obj = getattr(obj, attr)
        if is_decorated:
            return obj
        func = obj.func
    elif isinstance(func, bytes):
        func = loads(func)
    obj = cls(func, *args, **(kwargs or {}))
    obj.__dict__.update(userdict)
    return obj



import dask
import numpy as np
import pytest

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)
from xorq.conftest import array_types_df
from xorq.tests.util import assert_frame_equal, check_eq
from xorq.vendor import ibis


KEY_PREFIX = xo.config.options.cache.key_prefix


def test_duckdb_cache_parquet(con, parquet_dir, tmp_path):
    parquet_path = parquet_dir / "astronauts.parquet"
    expr = (
        xo.duckdb.connect()
        .read_parquet(parquet_path)[lambda t: t.number > 22]
        .cache(cache=ParquetCache.from_kwargs(source=con, relative_path=tmp_path))
    )
    expr.execute()


def test_duckdb_cache_csv(con, csv_dir, tmp_path):
    csv_path = csv_dir / "astronauts.csv"
    expr = (
        xo.duckdb.connect()
        .read_csv(csv_path)[lambda t: t.number > 22]
        .cache(cache=ParquetCache.from_kwargs(source=con, relative_path=tmp_path))
    )
    expr.execute()


def test_duckdb_cache_arrow(con, tmp_path):
    name = "astronauts"
    expr = (
        xo.duckdb.connect()
        .create_table(name, con.table(name).to_pyarrow())[lambda t: t.number > 22]
        .cache(cache=ParquetCache.from_kwargs(source=con, relative_path=tmp_path))
    )
    expr.execute()


def test_duckdb_snapshot(con_snapshot):
    con_snapshot(xo.duckdb.connect())


def test_cross_source_snapshot(con_cross_source_snapshot):
    con_cross_source_snapshot(xo.duckdb.connect())


@pytest.mark.parametrize(
    "cls",
    [ParquetSnapshotCache, ParquetCache, SourceSnapshotCache, SourceCache],
)
def test_cache_find_backend(cls, con_cache_find_backend):
    con_cache_find_backend(cls, xo.duckdb.connect())


def test_register_table_with_uppercase(con):
    db_con = xo.duckdb.connect()
    db_t = db_con.create_table("lowercase", schema=ibis.schema({"A": "int"}))

    uppercase_table_name = "UPPERCASE"
    t = con.register(db_t, uppercase_table_name)
    assert uppercase_table_name in con.list_tables()
    assert xo.execute(t) is not None


def test_register_table_with_uppercase_multiple_times(con):
    db_con = xo.duckdb.connect()
    db_t = db_con.create_table("lowercase", schema=ibis.schema({"A": "int"}))

    uppercase_table_name = "UPPERCASE"
    con.register(db_t, uppercase_table_name)

    expected_schema = ibis.schema({"B": "int"})
    db_t = db_con.create_table("lowercase_2", schema=expected_schema)
    t = con.register(db_t, uppercase_table_name)

    assert uppercase_table_name in con.list_tables()
    assert xo.execute(t) is not None
    assert t.schema() == expected_schema


def test_sql_execution(con, duckdb_con, awards_players, batting):
    def make_right(t):
        return t[t.lgID == "NL"].drop("yearID", "lgID")

    ddb_players = con.register(
        duckdb_con.table("ddb_players"), table_name="ddb_players"
    )

    left = batting[batting.yearID == 2015]
    right_df = make_right(awards_players).execute()
    left_df = xo.execute(left)
    predicate = ["playerID"]
    result_order = ["playerID", "yearID", "lgID", "stint"]

    expr = con.register(left, "batting").join(
        make_right(con.register(ddb_players, "players")),
        predicate,
        how="inner",
    )
    query = xo.to_sql(expr)

    result = (
        con.sql(query)
        .execute()
        .fillna(np.nan)[left.columns]
        .sort_values(result_order)
        .reset_index(drop=True)
    )

    expected = (
        check_eq(
            left_df,
            right_df,
            how="inner",
            on=predicate,
            suffixes=("_x", "_y"),
        )[left.columns]
        .sort_values(result_order)
        .reset_index(drop=True)
    )

    assert_frame_equal(result, expected, check_like=True)


def test_register_arbitrary_expression(con, duckdb_con):
    batting_table_name = "batting"
    t = duckdb_con.table(batting_table_name)

    expr = t.filter(t.playerID == "allisar01")[
        ["playerID", "yearID", "stint", "teamID", "lgID"]
    ]
    expected = expr.execute()

    ddb_batting_table_name = f"{duckdb_con.name}_{batting_table_name}"
    table = con.register(expr, table_name=ddb_batting_table_name)
    result = table.execute()

    assert result is not None
    assert_frame_equal(result, expected, check_like=True)


def test_array_filter_cached(con, duckdb_con):
    t = duckdb_con.create_table("array_types", array_types_df)
    uncached = t.mutate(filtered=t.x.filter(xo._ > 1)).into_backend(
        con, name="array_types"
    )
    expr = uncached.cache(SourceCache.from_kwargs(source=con))

    assert dask.base.tokenize(uncached)
    assert not expr.execute().empty



import xorq.vendor.ibis.expr.types as ir
from xorq.internal import AbstractTableProvider


class IbisTableProvider(AbstractTableProvider):
    def __init__(self, table: ir.Table):
        self.table = table

    def schema(self):
        return self.table.schema().to_pyarrow()

    def scan(self, filters=None):
        table = self.table
        if filters:
            table = self.table.filter(filters)
        backend = table._find_backend()
        return backend.to_pyarrow_batches(table)



import importlib

import dask
import toolz

import xorq.expr.relations as rel
import xorq.vendor.ibis.expr.operations as ops
from xorq.common.utils.graph_utils import (
    bfs,
    find_all_sources,
    replace_nodes,
    walk_nodes,
)
from xorq.vendor.ibis.expr.operations import UnboundTable


replace_typs = (
    ops.PhysicalTable,
    rel.CachedNode,
    rel.Read,
    # can't use Tag for anything involving hashes: hash value is the same as parent
    # rel.Tag,
)


def compute_expr_hash(expr, strategy=None):
    """
    Compute hash for an expression with optional normalization strategy.

    Args:
        expr: Expression to hash (can be Expr or node, will convert to expr if needed)
        strategy: Optional normalization strategy for consistent hashing

    Returns:
        Hash string for the expression
    """
    # Convert to expr if it's a node
    if hasattr(expr, "to_expr"):
        expr = expr.to_expr()

    if strategy is None:
        return expr.ls.tokenized

    with strategy.normalization_context(expr):
        return dask.base.tokenize(expr.ls.untagged)


@toolz.curry
def do_replace_dct(node, kwargs, *, replace_dct):
    if (replaced := replace_dct.get(node)) is not None:
        return replaced
    elif kwargs:
        return node.__recreate__(kwargs)
    else:
        return node


def get_typs(maybe_typs):
    match maybe_typs:
        case None:
            typs = replace_typs
        case tuple():
            typs = maybe_typs
        case str():
            (module, attr) = maybe_typs.rsplit(".", 1)
            typs = (getattr(importlib.import_module(module), attr),)
        case _:
            raise ValueError(
                f"unsupported maybe_typs type {type(maybe_typs)}: {maybe_typs!r}"
            )
    return typs


def find_by_expr_hash(expr, to_replace_hash, typs=None, strategy=None):
    def matches_hash(node):
        return compute_expr_hash(node, strategy) == to_replace_hash

    typs = get_typs(typs)
    (to_replace, *rest) = filter(
        matches_hash,
        walk_nodes(typs, expr),
    )
    if rest:
        raise ValueError(f"found multiple nodes with hash {to_replace_hash!r}")
    return to_replace


def find_by_expr_tag(expr, tag):
    yield from (node for node in walk_nodes(rel.Tag, expr) if node.tag == tag)


def find_node(expr, hash, tag, typs=None, strategy=None):
    match [hash, tag]:
        case [None, None]:
            raise ValueError("at least one of hash or tag must be provided")
        case [_, None]:
            if isinstance(typs, tuple) and rel.Tag in typs:
                raise ValueError("typs must not include Tag when searching by hash")
            return find_by_expr_hash(expr, hash, typs=typs, strategy=strategy)
        case [None, _]:
            (node, *rest) = find_by_expr_tag(expr, tag)
            if rest:
                raise ValueError(f"found multiple nodes with tag {tag!r}")
            else:
                return node
        case _:
            raise ValueError("cannot specify both hash and tag")


def replace_by_expr_hash(expr, to_replace_hash, replace_with, typs=None, strategy=None):
    typs = get_typs(typs)
    to_replace = find_by_expr_hash(expr, to_replace_hash, typs=typs, strategy=strategy)
    replaced = replace_nodes(
        do_replace_dct(
            replace_dct={to_replace: replace_with},
        ),
        expr,
    )
    return replaced.to_expr()


def unbind_expr_hash(expr, to_replace_hash, typs=None, strategy=None):
    typs = get_typs(typs)
    to_replace = find_by_expr_hash(expr, to_replace_hash, typs=typs, strategy=strategy)
    replace_with = ops.UnboundTable("unbound", to_replace.schema)
    unbound_expr = replace_nodes(
        do_replace_dct(
            replace_dct={to_replace: replace_with},
        ),
        expr,
    )
    return unbound_expr.to_expr()


def gen_downstream(expr, downstream_of):
    gi = bfs(expr).invert()
    visited = set()

    def inner(op):
        for child in gi[op]:
            if child not in visited:
                visited.add(child)
                yield child
                yield from inner(child)

    yield from inner(downstream_of)


def elide_downstream_cached_node(expr, downstream_of):
    cns = {
        el
        for el in gen_downstream(expr, downstream_of)
        if isinstance(el, rel.CachedNode)
    }

    def elide_cached_node(node, kwargs):
        if node in cns:
            while isinstance(node, rel.CachedNode):
                node = node.parent.op()
            node = replace_nodes(elide_cached_node, node.to_expr())
        if kwargs:
            node = node.__recreate__(kwargs)
        return node

    return elide_cached_node


def expr_to_unbound(expr, hash, tag, typs, strategy=None):
    """create an unbound expr that only needs to have a source of record batches fed in"""

    found = find_node(expr, hash=hash, tag=tag, typs=typs, strategy=strategy)
    found_expr = found.to_expr()
    to_unbind_hash = hash or compute_expr_hash(found_expr, strategy)
    found_con = None
    match find_all_sources(found_expr):
        case []:
            match walk_nodes(ops.InMemoryTable, found_expr):
                case []:
                    raise ValueError("found no connections")
                case [mt] if len(mt.data) == 0:
                    import warnings

                    import xorq.api as xo

                    warnings.warn(
                        "Found empty memtable, presuming partial_expr, using xo.connect()",
                        stacklevel=2,
                    )
                    found_con = xo.connect()
                case _:
                    raise ValueError(
                        "found multiple in-memory tables, cannot determine source"
                    )
            if not found_con:
                raise ValueError("found no connections")
        case [found_con]:
            pass
        case _:
            found_con = found_expr._find_backend()
            assert found_con
    unbound_table = UnboundTable("unbound", found.schema)
    replace_with = unbound_table.to_expr().into_backend(found_con).op()
    replaced = replace_by_expr_hash(
        expr, to_unbind_hash, replace_with, typs=(type(found),), strategy=strategy
    )
    (found,) = walk_nodes(UnboundTable, replaced)
    elided = replace_nodes(elide_downstream_cached_node(replaced, found), replaced)
    return elided


def recreate(op, **kwargs):
    new_kwargs = dict(zip(op.argnames, op.args)) | kwargs
    return op.__recreate__(new_kwargs)


def update_read_kwargs(old_read_kwargs, new_read_kwargs):
    existing = {name for name, _ in old_read_kwargs}
    to_append = tuple(
        (name, value) for (name, value) in new_read_kwargs if name not in existing
    )
    new_read_kwargs = dict(new_read_kwargs)
    modified_read_kwargs = (
        tuple(
            (name, new_read_kwargs.get(name, value)) for name, value in old_read_kwargs
        )
        + to_append
    )
    return modified_read_kwargs


def change_read_table_name(read, table_name):
    assert isinstance(read, rel.Read)
    read_kwargs = update_read_kwargs(read.read_kwargs, (("table_name", table_name),))
    return recreate(read, name=table_name, read_kwargs=read_kwargs)



from pathlib import Path
from shutil import move
from tarfile import TarFile
from tempfile import TemporaryDirectory
from urllib.request import urlretrieve

from xorq.init_templates import InitTemplates


def download_github_archive(org, repo, branch, suffix=".tar.gz", target=None):
    target = Path(target or f"{branch}{suffix}")
    assert not target.exists()
    archive_url = f"https://github.com/{org}/{repo}/archive/{branch}.tar.gz"
    _, _ = urlretrieve(archive_url, target)
    return target


def extract_tar(source, target):
    (source, target) = map(Path, (source, target))
    assert not target.exists()
    with TarFile.open(source, mode="r:*") as tf:
        (first, *rest) = (member.name for member in tf.members)
        assert all(member.startswith(first) for member in rest)
        with TemporaryDirectory() as td:
            tf.extractall(td)
            move(Path(td).joinpath(first), target)
    assert target.exists()
    return target


def download_xorq_template(template, branch=None, target=None):
    branch = branch or InitTemplates.get_default_branch(template)
    return download_github_archive(
        org="xorq-labs",
        repo=f"xorq-template-{template}",
        branch=branch,
        target=target,
    )


def download_unpacked_xorq_template(target, template, branch=None):
    target = Path(target)
    if target.exists():
        raise ValueError(
            f"download_unpacked_xorq_template: target `{target}` already exists"
        )
    with TemporaryDirectory() as td:
        archive_target = Path(td).joinpath("repo.tar.gz")
        download_xorq_template(
            template=template,
            branch=branch,
            target=archive_target,
        )
        extract_tar(archive_target, target)
        return target



import pytest
from pytest import param

from xorq.common.utils.ibis_utils import from_ibis


ibis = pytest.importorskip("ibis")


@pytest.mark.parametrize(
    "metrics",
    [
        (ibis._.int_col.first(),),
        (ibis._.bool_col.any(),),
        (ibis._.int_col.first(), ibis._.bool_col.any()),
    ],
)
def test_first(ibis_alltypes, metrics):
    expr = ibis_alltypes.group_by(ibis._.tinyint_col).aggregate(
        *metrics,
    )

    xo_expr = from_ibis(expr)
    assert xo_expr is not None
    assert not xo_expr.execute().empty


@pytest.mark.parametrize(
    "result_fn",
    [
        param(
            lambda t, where: t.bool_col.count(where=where),
            id="count",
        ),
        param(
            lambda t, where: t.bool_col.nunique(where=where),
            id="nunique",
        ),
        param(
            lambda t, where: t.bool_col.any(where=where),
            id="any",
        ),
        param(
            lambda t, where: t.bool_col.notany(where=where),
            id="notany",
        ),
        param(
            lambda t, where: -t.bool_col.any(where=where),
            id="any_negate",
        ),
        param(
            lambda t, where: t.bool_col.all(where=where),
            id="all",
        ),
        param(
            lambda t, where: t.bool_col.notall(where=where),
            id="notall",
        ),
        param(
            lambda t, where: -t.bool_col.all(where=where),
            id="all_negate",
        ),
        param(
            lambda t, where: t.double_col.sum(where=where),
            id="sum",
        ),
        param(
            lambda t, where: (t.int_col > 0).sum(where=where),
            id="bool_sum",
        ),
        param(
            lambda t, where: t.double_col.mean(where=where),
            id="mean",
        ),
        param(
            lambda t, where: t.double_col.min(where=where),
            id="min",
        ),
        param(
            lambda t, where: t.double_col.max(where=where),
            id="max",
        ),
        param(
            lambda t, where: t.double_col.std(how="sample", where=where),
            id="std",
        ),
        param(
            lambda t, where: t.double_col.var(how="sample", where=where),
            id="var",
        ),
        param(
            lambda t, where: t.double_col.std(how="pop", where=where),
            id="std_pop",
        ),
        param(
            lambda t, where: t.double_col.var(how="pop", where=where),
            id="var_pop",
        ),
        param(
            lambda t, where: t.string_col.nunique(where=where),
            id="string_nunique",
        ),
        param(
            lambda t, where: t.double_col.first(where=where),
            id="first",
        ),
        param(
            lambda t, where: t.double_col.last(where=where),
            id="last",
        ),
        param(
            lambda t, where: t.bigint_col.bit_and(where=where),
            id="bit_and",
        ),
        param(
            lambda t, where: t.bigint_col.bit_or(where=where),
            id="bit_or",
        ),
        param(
            lambda t, where: t.bigint_col.bit_xor(where=where),
            id="bit_xor",
        ),
        param(
            lambda t, where: t.count(where=where),
            id="count_star",
        ),
    ],
)
@pytest.mark.parametrize(
    "ibis_cond",
    [
        param(lambda _: None, id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
        param(
            lambda _: ibis._.string_col.isin(["1", "7"]),
            id="is_in_deferred",
        ),
    ],
)
def test_reduction_ops(
    ibis_alltypes,
    result_fn,
    ibis_cond,
):
    expr = ibis_alltypes.agg(tmp=result_fn(ibis_alltypes, ibis_cond(ibis_alltypes))).tmp

    xorq_expr = from_ibis(expr)
    assert not xorq_expr.execute().empty


def test_approx_median(ibis_alltypes):
    expr = ibis_alltypes.double_col.approx_median()

    xorq_expr = from_ibis(expr)
    assert xorq_expr.execute()


def test_median(ibis_alltypes):
    expr = ibis_alltypes.double_col.median()

    xorq_expr = from_ibis(expr)
    assert xorq_expr.execute()



import ibis

from xorq.common.utils.ibis_utils import from_ibis


def test_equals():
    a = ibis.literal(5)
    b = ibis.literal(5)
    expr = a == b

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_not_equals():
    a = ibis.literal(5)
    b = ibis.literal(3)
    expr = a != b

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_greater_than():
    a = ibis.literal(5)
    b = ibis.literal(3)
    expr = a > b

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_less_than():
    a = ibis.literal(3)
    b = ibis.literal(5)
    expr = a < b

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_and_or():
    a = ibis.literal(5)
    b = ibis.literal(3)
    c = ibis.literal(10)

    expr_and = (a > b) & (a < c)

    xorq_expr = from_ibis(expr_and)
    assert xorq_expr is not None


def test_not():
    a = ibis.literal(True)
    expr = ~a

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_is_null():
    a = ibis.literal(None)
    expr = a.isnull()

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_is_inf():
    a = ibis.literal(float("inf"))
    expr = a.isinf()

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_is_nan():
    from math import isnan

    a = ibis.literal(float("nan"))
    expr = a.isnan()

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None
    assert isnan(a.op().value)


def test_between():
    a = ibis.literal(5)
    expr = a.between(3, 7)

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None



import pytest

import xorq.api as xo
from xorq.common.utils.tls_utils import TLSKwargs
from xorq.flight import BasicAuth, FlightServer
from xorq.flight.tests.test_server import make_flight_url


@pytest.mark.parametrize("auth", [None, BasicAuth("username", "password")])
@pytest.mark.parametrize("verify_client", [False, True])
def test_connect(auth, verify_client):
    tls_kwargs = TLSKwargs.from_common_name(verify_client=verify_client)
    flight_url = make_flight_url(None, scheme="grpc+tls", auth=auth)

    with FlightServer(
        flight_url=flight_url,
        verify_client=verify_client,
        auth=auth,
        **tls_kwargs.server_kwargs,
    ) as _:
        con = xo.flight.connect(**flight_url.client_kwargs, tls_kwargs=tls_kwargs)
        assert con is not None
        assert con.list_exchanges()



"""Tests for ``xorq run --alias`` / ``--name``."""

import shutil
from pathlib import Path
from unittest.mock import patch

import click
import pytest
from click.testing import CliRunner

import xorq.api as xo
from xorq.catalog.catalog import Catalog
from xorq.catalog.expr_utils import build_expr_context_tgz
from xorq.cli import _resolve_alias, cli


# --- fixtures ---


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def catalog(tmp_path):
    repo = Catalog.init_repo_path(tmp_path / "catalog")
    return Catalog(repo=repo)


@pytest.fixture
def catalog_with_alias(catalog, tmp_path):
    expr = xo.memtable({"col": [1, 2, 3]})
    with build_expr_context_tgz(expr) as tgz_path:
        target = tmp_path / tgz_path.name
        shutil.copy(tgz_path, target)
        entry = catalog.add(target, aliases=("my-alias",))
    return catalog, entry


# --- validation tests ---


def test_run_no_args(runner):
    result = runner.invoke(cli, ["run"])
    assert result.exit_code != 0
    assert "Provide either BUILD_PATH or --alias" in result.output


def test_run_build_path_and_alias_mutually_exclusive(runner):
    result = runner.invoke(cli, ["run", "./some/path", "--alias", "foo"])
    assert result.exit_code != 0
    assert "mutually exclusive" in result.output


def test_run_name_without_alias(runner):
    result = runner.invoke(cli, ["run", "./some/path", "--name", "ns"])
    assert result.exit_code != 0
    assert "--name is only valid with --alias" in result.output


# --- alias resolution tests ---


def test_resolve_alias(catalog_with_alias):
    catalog, entry = catalog_with_alias
    with patch.object(Catalog, "from_kwargs", return_value=catalog):
        resolved = _resolve_alias("my-alias")
    assert resolved.name == entry.name


def test_resolve_alias_unknown(catalog_with_alias):
    catalog, _ = catalog_with_alias
    with patch.object(Catalog, "from_kwargs", return_value=catalog):
        with pytest.raises(click.ClickException, match="Unknown alias"):
            _resolve_alias("no-such")


def test_resolve_alias_lists_available(catalog_with_alias):
    catalog, _ = catalog_with_alias
    with patch.object(Catalog, "from_kwargs", return_value=catalog):
        with pytest.raises(click.ClickException, match="my-alias"):
            _resolve_alias("no-such")


def test_resolve_alias_with_name(catalog_with_alias):
    catalog, entry = catalog_with_alias
    with patch.object(Catalog, "from_kwargs", return_value=catalog) as mock:
        resolved = _resolve_alias("my-alias", name="my-ns")
    mock.assert_called_once_with(name="my-ns", init=False)
    assert resolved.name == entry.name


# --- integration tests ---


def test_run_alias_invokes_run_command(runner, catalog_with_alias):
    catalog, _ = catalog_with_alias
    with (
        patch.object(Catalog, "from_kwargs", return_value=catalog),
        patch("xorq.cli.run_command") as mock_run,
    ):
        result = runner.invoke(cli, ["run", "--alias", "my-alias"])
    assert result.exit_code == 0, result.output
    mock_run.assert_called_once()
    build_dir = mock_run.call_args[0][0]
    assert isinstance(build_dir, Path)


def test_run_alias_with_name_passes_name(runner, catalog_with_alias):
    catalog, _ = catalog_with_alias
    with (
        patch.object(Catalog, "from_kwargs", return_value=catalog) as mock_catalog,
        patch("xorq.cli.run_command") as mock_run,
    ):
        result = runner.invoke(cli, ["run", "--alias", "my-alias", "--name", "my-ns"])
    assert result.exit_code == 0, result.output
    mock_catalog.assert_called_once_with(name="my-ns", init=False)
    mock_run.assert_called_once()


def test_run_alias_forwards_options(runner, catalog_with_alias):
    catalog, _ = catalog_with_alias
    with (
        patch.object(Catalog, "from_kwargs", return_value=catalog),
        patch("xorq.cli.run_command") as mock_run,
    ):
        result = runner.invoke(
            cli,
            [
                "run",
                "--alias",
                "my-alias",
                "--format",
                "csv",
                "--limit",
                "10",
            ],
        )
    assert result.exit_code == 0, result.output
    _, output_path, output_format, cache_dir, limit = mock_run.call_args[0]
    assert output_format == "csv"
    assert limit == 10



import xorq.vendor.ibis as ibis
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_equals(compiler):
    a = ibis.literal(5)
    b = ibis.literal(5)
    expr = a == b
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Equals"
    assert expression["left"]["value"] == 5
    assert expression["right"]["value"] == 5
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert dtype_yaml == {
        "op": "DataType",
        "type": "Boolean",
        "nullable": {"op": "bool", "value": True},
    }
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_not_equals(compiler):
    a = ibis.literal(5)
    b = ibis.literal(3)
    expr = a != b
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "NotEquals"
    assert expression["left"]["value"] == 5
    assert expression["right"]["value"] == 3
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_greater_than(compiler):
    a = ibis.literal(5)
    b = ibis.literal(3)
    expr = a > b
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Greater"
    assert expression["left"]["value"] == 5
    assert expression["right"]["value"] == 3
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_less_than(compiler):
    a = ibis.literal(3)
    b = ibis.literal(5)
    expr = a < b
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Less"
    assert expression["left"]["value"] == 3
    assert expression["right"]["value"] == 5
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_and_or(compiler):
    a = ibis.literal(5)
    b = ibis.literal(3)
    c = ibis.literal(10)

    expr_and = (a > b) & (a < c)
    yaml_dict = compiler.to_yaml(expr_and)
    expression = yaml_dict["expression"]
    assert expression["op"] == "And"
    assert expression["left"]["op"] == "Greater"
    assert expression["right"]["op"] == "Less"
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr_and)

    expr_or = (a > b) | (a < c)
    yaml_dict = compiler.to_yaml(expr_or)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Or"
    assert expression["left"]["op"] == "Greater"
    assert expression["right"]["op"] == "Less"
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr_or)


def test_not(compiler):
    a = ibis.literal(True)
    expr = ~a
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Not"
    assert expression["arg"]["value"]
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_is_null(compiler):
    a = ibis.literal(None)
    expr = a.isnull()
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "IsNull"
    assert expression["arg"]["value"] is None
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_is_inf(compiler):
    a = ibis.literal(float("inf"))
    expr = a.isinf()
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "IsInf"
    assert expression["arg"]["value"] == float("inf")
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_is_nan(compiler):
    from math import isnan

    a = ibis.literal(float("nan"))
    expr = a.isnan()
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "IsNan"
    isnan(expression["arg"]["value"])
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_between(compiler):
    a = ibis.literal(5)
    expr = a.between(3, 7)
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Between"
    assert expression["arg"]["value"] == 5
    assert expression["lower_bound"]["value"] == 3
    assert expression["upper_bound"]["value"] == 7
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)



import inspect
from sysconfig import get_python_version


def get_enclosing_function(level=2):
    # let caller inspect it's caller's name with level=2
    return inspect.stack()[level].function


def maybe_unwrap_curry(func, *args, **kwargs):
    # from toolz.curry.__init__
    if (
        hasattr(func, "func")
        and hasattr(func, "args")
        and hasattr(func, "keywords")
        and isinstance(func.args, tuple)
    ):
        _kwargs = {}
        if func.keywords:
            _kwargs.update(func.keywords)
        _kwargs.update(kwargs)
        kwargs = _kwargs
        args = func.args + args
        func = func.func
    elif hasattr(func, "func"):
        func = func.func
    return (func, args, kwargs)


def get_arguments(f, *args, **kwargs):
    (f, args, kwargs) = maybe_unwrap_curry(f, *args, **kwargs)
    signature = inspect.signature(f)
    bound = signature.bind(*args, **kwargs)
    bound.apply_defaults()
    arguments = bound.arguments
    return arguments


def get_partial_arguments(f, *args, **kwargs):
    (f, args, kwargs) = maybe_unwrap_curry(f, *args, **kwargs)
    signature = inspect.signature(f)
    bound = signature.bind_partial(*args, **kwargs)
    bound.apply_defaults()
    arguments = bound.arguments
    return arguments


def get_python_version_no_dot():
    return get_python_version().replace(".", "")



import os
import pathlib
import sys
from pathlib import Path

from xorq.config import env_config
from xorq.expr.relations import CachedNode, Read
from xorq.vendor.ibis import BaseBackend
from xorq.vendor.ibis.expr import operations as ops


def get_xorq_cache_dir() -> pathlib.Path:
    # NOTE: modifying env var XORQ_CACHE_DIR won't have any impact after first import
    if path := env_config.XORQ_CACHE_DIR:
        return Path(path).expanduser()

    name = "xorq"
    if sys.platform == "win32":
        return Path(os.path.normpath(os.environ["LOCALAPPDATA"])).joinpath(
            name, "cache"
        )
    else:
        return Path(os.getenv("XDG_CACHE_HOME", f"~/.cache/{name}")).expanduser()


def find_backend(op: ops.Node, use_default=False) -> tuple[BaseBackend, bool]:
    backends = set()
    has_unbound = False
    node_types = (
        ops.UnboundTable,
        ops.DatabaseTable,
        ops.SQLQueryResult,
        CachedNode,
        Read,
    )
    for table in op.find(node_types):
        if isinstance(table, ops.UnboundTable):
            has_unbound = True
        else:
            backends.add(table.source)

    if not backends and use_default:
        from xorq.config import _backend_init

        con = _backend_init()
        backends.add(con)

    return (
        backends.pop(),
        has_unbound,
    )  # TODO what happens if it has more than one backend



import pytest

import xorq.vendor.ibis as ibis
from xorq.expr.relations import FlightUDXF


@pytest.mark.parametrize(
    "schema_val,pattern",
    [
        (ibis.schema({"x": "float64"}), "Schema validation failed, expected"),
        (None, "^Schema validation failed$"),
    ],
)
def test_flight_udxf_validate_schema_fail(schema_val, pattern):
    schema = ibis.schema({"x": "int64"})
    input_expr = ibis.table(schema, name="t")

    class DummyUDXF:
        schema_in_required = schema_val

        @staticmethod
        def schema_in_condition(sch):
            return False

        @staticmethod
        def calc_schema_out(sch):
            return None

    with pytest.raises(ValueError, match=pattern):
        FlightUDXF.validate_schema(input_expr, DummyUDXF)



"""Initialize xorq module."""

from __future__ import annotations

import sys

from xorq import examples
from xorq import udf
from xorq.udf import *  # noqa: F403
from xorq import config
from xorq import expr
from xorq.config import options

from xorq.expr import api
from xorq.expr.api import *  # noqa: F403
from xorq import caching
from xorq.caching import *  # noqa: F403
from xorq import ml
from xorq.ml import *  # noqa: F403
from xorq.backends.xorq import Backend
from xorq.internal import SessionConfig

from xorq.loader import load_backend
from xorq.ibis_yaml.compiler import (
    build_expr,
    load_expr,
)

__all__ = [  # noqa: PLE0604
    "api",
    "caching",
    "examples",
    "expr",
    "flight",
    "config",
    "connect",
    "options",
    "SessionConfig",
    "udf",
    "build_expr",
    "load_expr",
    *api.__all__,
    *caching.__all__,
    *ml.__all__,
    *udf.__all__,
]


def connect(session_config: SessionConfig | None = None) -> Backend:
    """Create a xorq backend."""
    instance = Backend()
    instance.do_connect(session_config)
    return instance


from xorq import flight  # noqa: E402


def __getattr__(name):
    from xorq.vendor import ibis

    backend = load_backend(name) or ibis.load_backend(name)

    setattr(sys.modules[__name__], name, backend)

    return backend



import pandas as pd
import pytest

import xorq.vendor.ibis as ibis
from xorq.common.utils.graph_utils import bfs, walk_nodes
from xorq.common.utils.node_utils import gen_downstream
from xorq.vendor.ibis.expr.operations import InMemoryTable


def create_expr(depth=3):
    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": [4, 5, 6],
            "c": [7, 8, 9],
        }
    )
    t = ibis.memtable(df)

    expr = t.mutate(x=t.a + t.b, y=t.b + t.c)

    for i in range(depth):
        expr = expr.mutate(
            **{
                f"d{i}": expr.x * 2 + expr.y,
                f"e{i}": expr.x + expr.y * 3,
                f"f{i}": expr.x * expr.y,
            }
        )

    return expr[[col for col in expr.columns if col.startswith(("d", "e", "f"))]][:5]


@pytest.mark.benchmark
@pytest.mark.parametrize("depth", [2, 3, 4])
def test_gen_downstream_performance(depth):
    expr = create_expr(depth)

    source = next(iter(walk_nodes(InMemoryTable, expr)))

    downstream = list(gen_downstream(expr, source))

    unique_nodes = len(bfs(expr).invert())
    visit_ratio = len(downstream) / unique_nodes

    assert visit_ratio <= 2.0, f"Expected linear visits, got {visit_ratio:.1f}x"



import pandas as pd
import pytest

import xorq.api as xo


@pytest.fixture(scope="function")
def con():
    return xo.connect()


@pytest.fixture(scope="session")
def iris_path() -> str:
    path = xo.options.pins.get_path("iris")
    return path


@pytest.fixture(scope="session")
def astronauts_parquet_path(parquet_dir):
    path = parquet_dir / "astronauts.parquet"
    return path


@pytest.fixture(scope="session")
def astronauts_df(astronauts_parquet_path) -> pd.DataFrame:
    return pd.read_parquet(astronauts_parquet_path)



import pytest

import xorq.api as xo
from xorq.examples import Example, get_name_to_suffix
from xorq.examples.core import whitelist


def test_whitelist():
    assert [name in dir(xo.examples) for name in whitelist]


def test_attributes():
    for name in whitelist:
        example = getattr(xo.examples, name)
        assert isinstance(example, Example)

    with pytest.raises(AttributeError):
        xo.examples.missing


@pytest.mark.parametrize(
    "fetch",
    [
        pytest.param(
            lambda: xo.examples.penguins.fetch(xo.duckdb.connect(), table_name="t"),
            id="duckdb-name",
        ),
        pytest.param(
            lambda: xo.examples.penguins.fetch(table_name="t"), id="default-name"
        ),
        pytest.param(lambda: xo.examples.penguins.fetch(), id="default-no-name"),
        pytest.param(
            lambda: xo.examples.penguins.fetch(deferred=False), id="default-no-deferred"
        ),
    ],
)
def test_fetch_penguins(fetch):
    t = fetch()
    expected = t.execute()

    assert not expected.empty
    assert expected.shape


def test_parquet_examples():
    parquet_examples = tuple(
        name for name, suffix in get_name_to_suffix().items() if suffix == ".parquet"
    )
    assert len(parquet_examples) > 0

    assert not any(
        getattr(xo.examples, example_name)
        .fetch(xo.duckdb.connect(), table_name="t")
        .execute()
        .empty
        for example_name in parquet_examples
    )


def test_fetch_default_backend():
    batting = xo.examples.batting.fetch()
    awards_players = xo.examples.awards_players.fetch()
    assert batting._find_backend() is awards_players._find_backend()
    on = set(batting.columns).intersection(awards_players.columns)
    joined = batting.select(on).join(awards_players.select(on), predicates=on).execute()
    assert not joined.empty



import functools
import importlib.metadata


@functools.cache
def _get_backend_names(*, exclude: tuple[str] = ()) -> frozenset[str]:
    """Return the set of known backend names.

    Parameters
    ----------
    exclude
        These backend names should be excluded from the result

    Notes
    -----
    This function returns a frozenset to prevent cache pollution.

    If a `set` is used, then any in-place modifications to the set
    are visible to every caller of this function.

    """

    entrypoints = importlib.metadata.entry_points(group="xorq.backends")
    return frozenset(ep.name for ep in entrypoints).difference(exclude)



import hashlib

import pytest

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.backends.snowflake.tests.conftest import inside_temp_schema
from xorq.tests.util import assert_series_equal
from xorq.vendor.ibis import literal as L


def calc_hexdigest(string, how):
    return hashlib.new(how, string.encode()).hexdigest()


@pytest.mark.snowflake
def test_create_table_from_expr_success(sf_con, temp_catalog, temp_db, parquet_dir):
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        name = "functional_alltypes"
        col_name, by = "string_col", "id"
        alltypes = (
            xo.deferred_read_parquet(
                parquet_dir.joinpath(f"{name}.parquet"), table_name=name
            )
            .limit(1000)
            .into_backend(sf_con)
        )

        expr = alltypes.order_by(by).limit(10)[col_name]
        for how in ("sha256", "sha512", "md5"):
            h1 = expr.hexdigest(how=how).execute()
            h2 = expr.execute().apply(calc_hexdigest, how=how).rename(h1.name)
            assert_series_equal(h1, h2)


@pytest.mark.parametrize(
    ("raw_value", "expected"), [("a", 0), ("b", 1), ("d", -1), (None, 3)]
)
def test_find_in_set(sf_con, raw_value, expected):
    value = L(raw_value, dt.string)
    haystack = ["a", "b", "c", None]
    expr = value.find_in_set(haystack)
    assert sf_con.execute(expr) == expected


def test_string_column_find_in_set(sf_con, temp_catalog, temp_db, parquet_dir):
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        name = "functional_alltypes"
        alltypes = (
            xo.deferred_read_parquet(
                parquet_dir.joinpath(f"{name}.parquet"), table_name=name
            )
            .limit(1000)
            .into_backend(sf_con)
        )

        s = alltypes.string_col
        vals = "312"

        actual = s.find_in_set(list(vals)).execute().to_list()
        expected = [vals.find(v) for v in s.execute()]

        assert actual == expected



from __future__ import annotations

import base64
import datetime
import decimal
import functools
import warnings
from collections.abc import Callable
from pathlib import Path
from typing import Any

import dask
import toolz

import xorq.expr.datatypes as dt
import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.datashape as ds
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.operations.temporal as tm
import xorq.vendor.ibis.expr.types as ir
from xorq.common.utils.name_utils import get_uid_prefix
from xorq.common.utils.node_utils import update_read_kwargs
from xorq.expr.relations import (
    CachedNode,
    Read,
    RemoteTable,
    Tag,
    into_backend,
)
from xorq.ibis_yaml.common import (
    RefEnum,
    RegistryEnum,
    TranslationContext,
    _translate_type,
    deserialize_callable,
    register_from_yaml_handler,
    serialize_callable,
    translate_to_yaml,
)
from xorq.ibis_yaml.udf import _scalar_udf_from_yaml, _scalar_udf_to_yaml  # noqa: F401
from xorq.ibis_yaml.utils import (
    freeze,
    load_cache_from_yaml,
    translate_cache,
)
from xorq.vendor.ibis.common.collections import FrozenDict, FrozenOrderedDict
from xorq.vendor.ibis.expr.datashape import Columnar
from xorq.vendor.ibis.expr.operations.relations import Namespace
from xorq.vendor.ibis.expr.schema import Schema
from xorq.vendor.ibis.util import normalize_filenames


@toolz.curry
def convert_to_ref(which, wrapped):
    @functools.wraps(wrapped)
    def wrapper(op, context):
        frozen = wrapped(op, context)
        if context is None:
            return frozen
        else:
            ref = context.register(which, op, frozen)
            return ref

    return wrapper


convert_to_dtype_ref = convert_to_ref(RegistryEnum.dtypes)
convert_to_node_ref = convert_to_ref(RegistryEnum.nodes)
convert_to_schema_ref = convert_to_ref(RegistryEnum.schemas)


@translate_to_yaml.register(ops.Node)
def _object_to_yaml(obj: ops.Node, context: Any) -> dict:
    return freeze(
        {"op": obj.__class__.__name__}
        | {
            name: context.translate_to_yaml(arg)
            for name, arg in zip(obj.argnames, obj.args)
        }
        | {
            name: context.translate_to_yaml(getattr(obj, attribute))
            for name, attribute in (("type", "dtype"),)
            if hasattr(obj, attribute)
        }
    )


@translate_to_yaml.register(Callable)
def _callable_to_yaml(obj: Callable, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Callable",
            "pickled_fn": serialize_callable(obj),
        }
    )


@register_from_yaml_handler("Callable")
def _callable_from_yaml(yaml_dict: dict, compiler: Any) -> Callable:
    return deserialize_callable(yaml_dict["pickled_fn"])


@translate_to_yaml.register(tuple)
def _tuple_to_yaml(tpl: tuple, context: TranslationContext) -> dict:
    return freeze(
        {"op": "tuple", "values": [context.translate_to_yaml(value) for value in tpl]}
    )


@register_from_yaml_handler("tuple")
def _tuple_from_yaml(yaml_dict: dict, context: TranslationContext) -> Any:
    return tuple(context.translate_from_yaml(value) for value in yaml_dict["values"])


@translate_to_yaml.register(frozenset)
def _frozenset_to_yaml(tpl: tuple, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "frozenset",
            "values": [context.translate_to_yaml(value) for value in tpl],
        }
    )


@register_from_yaml_handler("frozenset")
def _frozenset_from_yaml(yaml_dict: dict, context: TranslationContext) -> Any:
    return frozenset(
        context.translate_from_yaml(value) for value in yaml_dict["values"]
    )


@translate_to_yaml.register(tm.IntervalUnit)
def _interval_unit_to_yaml(
    interval_unit: tm.IntervalUnit, context: TranslationContext
) -> dict:
    if interval_unit.is_date():
        unit_name = "DateUnit"
    elif interval_unit.is_time():
        unit_name = "TimeUnit"
    else:
        unit_name = "IntervalUnit"

    return freeze(
        {"op": "IntervalUnit", "name": unit_name, "value": interval_unit.value}
    )


@translate_to_yaml.register(tm.DateUnit)
def _date_unit_to_yaml(unit: tm.DateUnit, context: TranslationContext) -> dict:
    return freeze({"op": "IntervalUnit", "name": "DateUnit", "value": unit.value})


@translate_to_yaml.register(tm.TimeUnit)
def _time_unit_to_yaml(unit: tm.TimeUnit, context: TranslationContext) -> dict:
    return freeze({"op": "IntervalUnit", "name": "TimeUnit", "value": unit.value})


@translate_to_yaml.register(tm.TimestampUnit)
def _timestamp_unit_to_yaml(
    unit: tm.TimestampUnit, context: TranslationContext
) -> dict:
    return freeze({"op": "IntervalUnit", "name": "TimestampUnit", "value": unit.value})


@register_from_yaml_handler("IntervalUnit")
def _interval_unit_from_yaml(yaml_dict: dict, context: TranslationContext) -> any:
    unit_classes = {
        "DateUnit": tm.DateUnit,
        "TimeUnit": tm.TimeUnit,
        "TimestampUnit": tm.TimestampUnit,
        "IntervalUnit": tm.IntervalUnit,
    }
    name = yaml_dict.get("name", "IntervalUnit")
    cls = unit_classes.get(name, tm.IntervalUnit)
    return cls(yaml_dict["value"])


@translate_to_yaml.register(bool)
def _bool_to_yaml(value: bool, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "bool",
            "value": value,
        }
    )


@register_from_yaml_handler("bool")
def _bool_from_yaml(yaml_dict: dict, context: TranslationContext) -> bool:
    return yaml_dict["value"]


@translate_to_yaml.register(int)
def _int_to_yaml(val: int, context: TranslationContext) -> dict:
    return freeze({"op": "int", "value": val})


@register_from_yaml_handler("int")
def _int_from_yaml(yaml_dict: dict, context: TranslationContext) -> any:
    return int(yaml_dict["value"])


@translate_to_yaml.register(float)
def _float_to_yaml(dct: float, context: TranslationContext) -> dict:
    return freeze({"op": "float", "value": str(dct)})


@register_from_yaml_handler("float")
def _float_from_yaml(yaml_dict: dict, context: TranslationContext) -> any:
    return float(yaml_dict["value"])


@_translate_type.register(dt.Timestamp)
def _translate_timestamp_type(dtype: dt.Timestamp) -> dict:
    base = {"name": "Timestamp", "nullable": dtype.nullable}
    if dtype.timezone is not None:
        base["timezone"] = dtype.timezone
    return freeze(base)


@_translate_type.register(dt.Decimal)
def _translate_decimal_type(dtype: dt.Decimal) -> dict:
    base = {"name": "Decimal", "nullable": dtype.nullable}
    if dtype.precision is not None:
        base["precision"] = dtype.precision
    if dtype.scale is not None:
        base["scale"] = dtype.scale
    return freeze(base)


@_translate_type.register(dt.Array)
def _translate_array_type(dtype: dt.Array) -> dict:
    return freeze(
        {
            "name": "Array",
            "value_type": translate_to_yaml(dtype.value_type, None),
            "nullable": dtype.nullable,
        }
    )


@_translate_type.register(dt.Map)
def _translate_map_type(dtype: dt.Map) -> dict:
    return freeze(
        {
            "name": "Map",
            "key_type": translate_to_yaml(dtype.key_type, None),
            "value_type": translate_to_yaml(dtype.value_type, None),
            "nullable": dtype.nullable,
        }
    )


@_translate_type.register(dt.Interval)
def _translate_type_interval(dtype: dt.Interval) -> dict:
    return freeze(
        {
            "name": "Interval",
            "unit": _translate_temporal_unit(dtype.unit),
            "nullable": dtype.nullable,
        }
    )


@_translate_type.register(dt.Struct)
def _translate_struct_type(dtype: dt.Struct) -> dict:
    return freeze(
        {
            "name": "Struct",
            "fields": {
                name: _translate_type(field_type)
                for name, field_type in zip(dtype.names, dtype.types)
            },
            "nullable": dtype.nullable,
        }
    )


def _translate_temporal_unit(unit: tm.IntervalUnit) -> dict:
    if unit.is_date():
        unit_name = "DateUnit"
    elif unit.is_time():
        unit_name = "TimeUnit"
    else:
        unit_name = "IntervalUnit"
    return freeze({"name": unit_name, "value": unit.value})


def _translate_literal_value(value: Any, dtype: dt.DataType) -> Any:
    if value is None:
        return None
    elif isinstance(value, (bool, int, float, str)):
        return value
    elif isinstance(value, decimal.Decimal):
        return str(value)
    elif isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        return value.isoformat()
    elif isinstance(value, list):
        return [_translate_literal_value(v, dtype.value_type) for v in value]
    elif isinstance(value, dict):
        if isinstance(dtype, dt.Struct):
            raise NotImplementedError
        else:
            return {
                _translate_literal_value(k, dtype.key_type): _translate_literal_value(
                    v, dtype.value_type
                )
                for k, v in value.items()
            }
    else:
        return value


@translate_to_yaml.register(dt.DataType)
@convert_to_dtype_ref
def _datatype_to_yaml(dtype: dt.DataType, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "DataType",
            "type": type(dtype).__name__,
        }
        | {
            argname: context.translate_to_yaml(arg)
            if context is not None
            else translate_to_yaml(arg, context)
            for argname, arg in zip(dtype.argnames, dtype.args)
        }
    )


@register_from_yaml_handler("DataType")
def _datatype_from_yaml(yaml_dict: dict, context: TranslationContext) -> any:
    typ = getattr(dt, yaml_dict["type"])
    dct = toolz.dissoc(yaml_dict, "op", "type")
    return typ(
        **{
            key: context.translate_from_yaml(value) if value is not None else None
            for key, value in dct.items()
        }
    )


@translate_to_yaml.register(ir.Expr)
def _expr_to_yaml(expr: ir.Expr, context: any) -> dict:
    return context.translate_to_yaml(expr.op())


@translate_to_yaml.register(ops.WindowFunction)
def _window_function_to_yaml(
    op: ops.WindowFunction, context: TranslationContext
) -> dict:
    result = {
        "op": "WindowFunction",
        "args": [context.translate_to_yaml(op.func)],
        "type": context.translate_to_yaml(op.dtype),
    }

    if op.group_by:
        result["group_by"] = [context.translate_to_yaml(expr) for expr in op.group_by]

    if op.order_by:
        result["order_by"] = [context.translate_to_yaml(expr) for expr in op.order_by]

    if op.start is not None:
        result["start"] = (
            context.translate_to_yaml(op.start.value)["value"]
            if isinstance(op.start, ops.WindowBoundary)
            else op.start
        )

    if op.end is not None:
        result["end"] = (
            context.translate_to_yaml(op.end.value)["value"]
            if isinstance(op.end, ops.WindowBoundary)
            else op.end
        )

    return freeze(result)


@register_from_yaml_handler("WindowFunction")
def _window_function_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    func = context.translate_from_yaml(yaml_dict["args"][0])
    group_by = [context.translate_from_yaml(g) for g in yaml_dict.get("group_by", [])]
    order_by = [context.translate_from_yaml(o) for o in yaml_dict.get("order_by", [])]
    start = ibis.literal(yaml_dict["start"]) if "start" in yaml_dict else None
    end = ibis.literal(yaml_dict["end"]) if "end" in yaml_dict else None
    window = ibis.window(
        group_by=group_by, order_by=order_by, preceding=start, following=end
    )
    return func.over(window)


@translate_to_yaml.register(ops.WindowBoundary)
def _window_boundary_to_yaml(
    op: ops.WindowBoundary, context: TranslationContext
) -> dict:
    return freeze(
        {
            "op": "WindowBoundary",
            "value": context.translate_to_yaml(op.value),
            "preceding": op.preceding,
            "type": context.translate_to_yaml(op.dtype),
        }
    )


@register_from_yaml_handler("WindowBoundary")
def _window_boundary_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    value = context.translate_from_yaml(yaml_dict["value"])
    return ops.WindowBoundary(value, preceding=yaml_dict["preceding"])


@translate_to_yaml.register(ops.StructField)
def _struct_field_to_yaml(op: ops.StructField, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "args": [context.translate_to_yaml(arg) for arg in op.args],
        }
    )


@translate_to_yaml.register(ops.UnboundTable)
@convert_to_node_ref
def _unbound_table_to_yaml(op: ops.UnboundTable, context: TranslationContext) -> dict:
    namespace_dict = freeze(
        {
            "catalog": op.namespace.catalog,
            "database": op.namespace.database,
        }
    )
    return freeze(
        {
            "op": "UnboundTable",
            "name": op.name,
            "namespace": namespace_dict,
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("UnboundTable")
def _unbound_table_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    table_name = yaml_dict["name"]
    namespace_dict = yaml_dict.get("namespace", {})
    catalog = namespace_dict.get("catalog")
    database = namespace_dict.get("database")
    schema = context.get_schema(yaml_dict[RefEnum.schema_ref])
    # TODO: use UnboundTable node to construct instead of builder API
    return ibis.table(schema, name=table_name, catalog=catalog, database=database)


@translate_to_yaml.register(ops.DatabaseTable)
@convert_to_node_ref
def _database_table_to_yaml(op: ops.DatabaseTable, context: TranslationContext) -> dict:
    profile_name = op.source._profile.hash_name
    namespace_dict = freeze(
        {
            "catalog": op.namespace.catalog,
            "database": op.namespace.database,
        }
    )

    node_dict = freeze(
        {
            "op": "DatabaseTable",
            "table": op.name,
            "profile": profile_name,
            "namespace": namespace_dict,
        }
        | context.registry.register_schema(op.schema)
    )
    return node_dict


@register_from_yaml_handler("DatabaseTable")
def database_table_from_yaml(yaml_dict: dict, context: TranslationContext) -> ibis.Expr:
    profile_name = yaml_dict.get("profile")
    table_name = yaml_dict.get("table")
    namespace_dict = yaml_dict.get("namespace", {})
    catalog = namespace_dict.get("catalog")
    database = namespace_dict.get("database")
    # we should validate that schema is the same
    schema = context.get_schema(yaml_dict.get(RefEnum.schema_ref))

    try:
        con = context.profiles[profile_name]
    except KeyError as err:
        raise ValueError(
            f"Profile {profile_name!r} not found in context.profiles"
        ) from err
    return ops.DatabaseTable(
        schema=schema,
        source=con,
        name=table_name,
        namespace=Namespace(catalog=catalog, database=database),
    ).to_expr()


@translate_to_yaml.register(CachedNode)
@convert_to_node_ref
def _cached_node_to_yaml(op: CachedNode, context: any) -> dict:
    # source should be called profile_name
    return freeze(
        {
            "op": "CachedNode",
            "name": op.name,
            "parent": context.translate_to_yaml(op.parent),
            "source": op.source._profile.hash_name,
            "cache": translate_cache(op.cache, context),
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("CachedNode")
def _cached_node_from_yaml(yaml_dict: dict, context: any) -> ibis.Expr:
    schema = context.get_schema(yaml_dict[RefEnum.schema_ref])
    name = yaml_dict["name"]

    parent_expr = context.translate_from_yaml(yaml_dict["parent"])
    profile_name = yaml_dict.get("source")
    try:
        source = context.profiles[profile_name]
    except KeyError as err:
        raise ValueError(
            f"Profile {profile_name!r} not found in context.profiles"
        ) from err
    cache = load_cache_from_yaml(yaml_dict["cache"], context)

    op = CachedNode(
        schema=schema,
        name=name,
        parent=parent_expr,
        source=source,
        cache=cache,
    )
    return op.to_expr()


@translate_to_yaml.register(RemoteTable)
@convert_to_node_ref
def _remotetable_to_yaml(op: RemoteTable, context: TranslationContext) -> dict:
    deterministic_name = dask.base.tokenize(op)
    profile_name = op.source._profile.hash_name
    remote_expr_yaml = context.translate_to_yaml(op.remote_expr)
    return freeze(
        {
            "op": "RemoteTable",
            "table": deterministic_name,
            "profile": profile_name,
            "remote_expr": remote_expr_yaml,
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("RemoteTable")
def _remotetable_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    profile_name = yaml_dict.get("profile")
    table_name = yaml_dict.get("table")
    remote_expr_yaml = yaml_dict.get("remote_expr")
    if profile_name is None:
        raise ValueError(
            "Missing keys in RemoteTable YAML; ensure 'profile_name' are present."
        )
    try:
        con = context.profiles[profile_name]
    except KeyError as err:
        raise ValueError(
            f"Profile {profile_name!r} not found in context.profiles"
        ) from err

    remote_expr = context.translate_from_yaml(remote_expr_yaml)

    remote_table_expr = into_backend(remote_expr, con, table_name)
    return remote_table_expr


def warn_on_local_path(items: dict) -> None:
    from urllib.parse import urlparse

    def is_local_path(any: str | Path) -> bool:
        parsed = urlparse(any)
        return not parsed.scheme or parsed.scheme == "file"

    if path := next(
        (v for k, v in dict(items).items() if k in ("path", "source")), None
    ):
        f = toolz.excepts((ValueError, AttributeError), is_local_path)
        paths = normalize_filenames(path)
        if any(map(f, paths)):
            warnings.warn(
                "The Read op path is using a local filesystem path, running the build may not work in other environments.",
                stacklevel=2,
            )


@translate_to_yaml.register(Read)
@convert_to_node_ref
def _read_to_yaml(op: Read, context: TranslationContext) -> dict:
    profile_hash_name = (
        op.source._profile.hash_name if hasattr(op.source, "_profile") else None
    )

    warn_on_local_path(op.read_kwargs)

    table_name = op.name
    read_kwargs = freeze(op.read_kwargs)
    if prefix := get_uid_prefix(table_name):
        table_name = f"{prefix}{dask.base.tokenize(op)}"
        read_kwargs = update_read_kwargs(read_kwargs, (("table_name", table_name),))
    return freeze(
        {
            "op": "Read",
            "method_name": op.method_name,
            "name": table_name,
            "profile": profile_hash_name,
            "read_kwargs": read_kwargs,
            "normalize_method": serialize_callable(op.normalize_method),
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("Read")
def _read_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    schema = context.get_schema(yaml_dict[RefEnum.schema_ref])
    source = context.profiles[yaml_dict["profile"]]
    read_kwargs = tuple(
        (k, ibis.schema(v)) if k == "schema" else (k, v)
        for k, v in yaml_dict.get("read_kwargs", ())
    )
    read_op = Read(
        method_name=yaml_dict["method_name"],
        name=yaml_dict["name"],
        schema=schema,
        source=source,
        read_kwargs=read_kwargs,
        normalize_method=deserialize_callable(yaml_dict["normalize_method"]),
    )

    return read_op.to_expr()


@translate_to_yaml.register(str)
def _str_to_yaml(string: str, context: TranslationContext) -> str:
    return {
        "op": "str",
        "value": string,
    }


@register_from_yaml_handler("str")
def _str_from_yaml(yaml_dict: dict, context: TranslationContext) -> str:
    return yaml_dict["value"]


@translate_to_yaml.register(bytes)
def _bytes_to_yaml(value: bytes, context: TranslationContext) -> dict:
    return {
        "op": "bytes",
        "value": base64.b64encode(value).decode("ascii"),
    }


@register_from_yaml_handler("bytes")
def _bytes_from_yaml(yaml_dict: dict, context: TranslationContext) -> bytes:
    return base64.b64decode(yaml_dict["value"])


@translate_to_yaml.register(ops.Alias)
def _alias_to_yaml(op: ops.Alias, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Alias",
            "type": context.translate_to_yaml(op.dtype),
            "args": [context.translate_to_yaml(arg) for arg in op.args],
        }
    )


@register_from_yaml_handler("Alias")
def _alias_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    (arg, name) = (context.translate_from_yaml(arg) for arg in yaml_dict["args"])
    return arg.name(name)


@translate_to_yaml.register(ops.Round)
def _round_to_yaml(op: ops.Round, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": op.__class__.__name__,
            "type": context.translate_to_yaml(op.dtype),
            "args": [context.translate_to_yaml(arg) for arg in op.args],
        }
    )


@register_from_yaml_handler("Round")
def _round_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    (arg, digits) = (
        None if arg is None else context.translate_from_yaml(arg)
        for arg in yaml_dict["args"]
    )
    return arg.round(digits)


@translate_to_yaml.register(type(None))
def _none_to_yaml(value: None, context: TranslationContext) -> None:
    return None


@translate_to_yaml.register(ops.Literal)
def _literal_to_yaml(op: ops.Literal, context: TranslationContext) -> dict:
    value = _translate_literal_value(op.value, op.dtype)
    return freeze(
        {"op": "Literal", "value": value, "type": context.translate_to_yaml(op.dtype)}
    )


@register_from_yaml_handler("Literal")
def _literal_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    value = yaml_dict["value"]
    dtype = context.translate_from_yaml(yaml_dict["type"])
    return ibis.literal(value, type=dtype)


@translate_to_yaml.register(ops.Lag)
def _lag_to_yaml(op: ops.Lag, context: TranslationContext) -> dict:
    result = {
        "op": "Lag",
        "arg": context.translate_to_yaml(op.arg),
        "type": context.translate_to_yaml(op.dtype),
    }

    if op.offset is not None:
        result["offset"] = context.translate_to_yaml(op.offset)

    if op.default is not None:
        result["default"] = context.translate_to_yaml(op.default)

    node_dict = freeze(result)

    return node_dict


@register_from_yaml_handler("Lag")
def _lag_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = context.translate_from_yaml(yaml_dict["arg"])

    offset = None
    if "offset" in yaml_dict:
        offset = context.translate_from_yaml(yaml_dict["offset"])

    default = None
    if "default" in yaml_dict:
        default = context.translate_from_yaml(yaml_dict["default"])

    return arg.lag(offset, default)


@translate_to_yaml.register(ops.Intersection)
@translate_to_yaml.register(ops.Union)
@translate_to_yaml.register(ops.Difference)
def _set_op_to_yaml(op: ops.Set, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "left": context.translate_to_yaml(op.left),
            "right": context.translate_to_yaml(op.right),
            "distinct": context.translate_to_yaml(op.distinct),
            "values": {
                name: context.translate_to_yaml(val) for name, val in op.values.items()
            },
        }
    )


set_ops_map = {
    "Intersection": ops.Intersection,
    "Union": ops.Union,
    "Difference": ops.Difference,
}


@register_from_yaml_handler("Intersection", "Union", "Difference")
def _set_op(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    left = context.translate_from_yaml(yaml_dict["left"])
    right = context.translate_from_yaml(yaml_dict["right"])
    distinct = context.translate_from_yaml(yaml_dict["distinct"])
    set_op = set_ops_map[yaml_dict["op"]]
    return set_op(left, right, distinct=distinct).to_expr()


@translate_to_yaml.register(ops.Binary)
def _binary_op_to_yaml(op: ops.Binary, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "left": context.translate_to_yaml(op.left),
            "right": context.translate_to_yaml(op.right),
            "type": context.translate_to_yaml(op.dtype),
        }
    )


@translate_to_yaml.register(ops.Filter)
@convert_to_node_ref
def _filter_to_yaml(op: ops.Filter, context: TranslationContext) -> dict:
    node_dict = freeze(
        {
            "op": "Filter",
            "parent": context.translate_to_yaml(op.parent),
            "predicates": [context.translate_to_yaml(pred) for pred in op.predicates],
        }
    )
    return node_dict


@register_from_yaml_handler("Filter")
def _filter_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = context.translate_from_yaml(yaml_dict["parent"])
    predicates = [context.translate_from_yaml(pred) for pred in yaml_dict["predicates"]]
    filter_op = ops.Filter(parent, predicates)
    return filter_op.to_expr()


@translate_to_yaml.register(ops.Project)
@convert_to_node_ref
def _project_to_yaml(op: ops.Project, context: TranslationContext) -> dict:
    node_dict = {
        "op": "Project",
        "parent": context.translate_to_yaml(op.parent),
        "values": {
            name: context.translate_to_yaml(val) for name, val in op.values.items()
        },
    }
    return freeze(node_dict)


@register_from_yaml_handler("Project")
def _project_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = context.translate_from_yaml(yaml_dict["parent"])

    values_dict = yaml_dict.get("values", {})

    values = {
        name: context.translate_from_yaml(val) for name, val in values_dict.items()
    }

    projected = parent.projection(values)
    return projected


@translate_to_yaml.register(ops.Aggregate)
def _aggregate_to_yaml(op: ops.Aggregate, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Aggregate",
            "parent": context.translate_to_yaml(op.parent),
            "by": {
                name: context.translate_to_yaml(group)
                for name, group in op.groups.items()
            },
            "metrics": {
                name: context.translate_to_yaml(metric)
                for name, metric in op.metrics.items()
            },
        }
    )


@register_from_yaml_handler("Aggregate")
def _aggregate_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = context.translate_from_yaml(yaml_dict["parent"])
    by = yaml_dict.get("by", {})
    if isinstance(by, dict):
        group_mapping = {
            name: context.translate_from_yaml(expr) for name, expr in by.items()
        }
    else:
        exprs = [context.translate_from_yaml(expr) for expr in by]
        group_mapping = {expr.get_name(): expr for expr in exprs}

    metrics = {
        name: context.translate_from_yaml(metric)
        for name, metric in yaml_dict.get("metrics", {}).items()
    }

    if group_mapping:
        result = parent.group_by(list(group_mapping.values())).aggregate(metrics)
        rename_map = {
            alias: expr.get_name()
            for alias, expr in group_mapping.items()
            if alias != expr.get_name()
        }
        if rename_map:
            result = result.rename(rename_map)
    else:
        result = parent.aggregate(metrics)

    return result


@translate_to_yaml.register(ops.JoinChain)
@convert_to_node_ref
def _join_to_yaml(op: ops.JoinChain, context: TranslationContext) -> dict:
    node_dict = {
        "op": "JoinChain",
        "first": context.translate_to_yaml(op.first),
        "rest": [
            {
                "how": link.how,
                "table": context.translate_to_yaml(link.table),
                "predicates": [
                    context.translate_to_yaml(pred) for pred in link.predicates
                ],
            }
            for link in op.rest
        ],
        "values": {
            name: context.translate_to_yaml(val) for name, val in op.values.items()
        },
    }
    return freeze(node_dict)


@register_from_yaml_handler("JoinChain")
def _join_chain_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    first = context.translate_from_yaml(yaml_dict["first"])
    result = first

    for join in yaml_dict["rest"]:
        table = context.translate_from_yaml(join["table"])
        predicates = [context.translate_from_yaml(pred) for pred in join["predicates"]]
        result = result.join(table, predicates, how=join["how"])

    values = {
        name: context.translate_from_yaml(val)
        for name, val in yaml_dict["values"].items()
    }
    result = result.select(values)
    return result


@translate_to_yaml.register(ops.ScalarSubquery)
def _scalar_subquery_to_yaml(
    op: ops.ScalarSubquery, context: TranslationContext
) -> dict:
    return freeze(
        {
            "op": "ScalarSubquery",
            "args": [context.translate_to_yaml(arg) for arg in op.args],
            "type": context.translate_to_yaml(op.dtype),
        }
    )


@register_from_yaml_handler("ScalarSubquery")
def _scalar_subquery_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    subquery = context.translate_from_yaml(yaml_dict["args"][0])
    return ops.ScalarSubquery(subquery).to_expr()


@translate_to_yaml.register(ops.ExistsSubquery)
def _exists_subquery_to_yaml(
    op: ops.ExistsSubquery, context: TranslationContext
) -> dict:
    return freeze(
        {
            "op": "ExistsSubquery",
            "rel": context.translate_to_yaml(op.rel),
            "type": context.translate_to_yaml(op.dtype),
        }
    )


@register_from_yaml_handler("ExistsSubquery")
def _exists_subquery_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    rel = context.translate_from_yaml(yaml_dict["rel"])
    return ops.ExistsSubquery(rel).to_expr()


@translate_to_yaml.register(ops.InSubquery)
def _in_subquery_to_yaml(op: ops.InSubquery, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "InSubquery",
            "needle": context.translate_to_yaml(op.needle),
            "haystack": context.translate_to_yaml(op.rel),
            "type": context.translate_to_yaml(op.dtype),
        }
    )


@register_from_yaml_handler("InSubquery")
def _in_subquery_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    needle = context.translate_from_yaml(yaml_dict["needle"])
    haystack = context.translate_from_yaml(yaml_dict["haystack"])
    return ops.InSubquery(haystack, needle).to_expr()


@translate_to_yaml.register(ops.Field)
def _field_to_yaml(op: ops.Field, context: TranslationContext) -> dict:
    result = {
        "op": "Field",
        "name": op.name,
        "relation": context.translate_to_yaml(op.rel),
        "type": context.translate_to_yaml(op.dtype),
    }

    if op.args and len(op.args) >= 2 and isinstance(op.args[1], str):
        underlying_name = op.args[1]
        if underlying_name != op.name:
            result["original_name"] = underlying_name

    node_dict = freeze(result)
    return node_dict


@register_from_yaml_handler("Field")
def field_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    relation = context.translate_from_yaml(yaml_dict["relation"])
    target_name = yaml_dict["name"]
    source_name = yaml_dict.get("original_name", target_name)
    field = relation[source_name]
    if target_name != source_name:
        field = field.name(target_name)

    return freeze(field)


@translate_to_yaml.register(ops.SelfReference)
def _self_reference_to_yaml(op: ops.SelfReference, context: TranslationContext) -> dict:
    result = {"op": "SelfReference", "identifier": op.identifier}
    if op.args:
        result["args"] = [context.translate_to_yaml(op.args[0])]
    return freeze(result)


@register_from_yaml_handler("SelfReference")
def _self_reference_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    underlying = None
    if "args" in yaml_dict and yaml_dict["args"]:
        underlying = context.translate_from_yaml(yaml_dict["args"][0])
    else:
        if underlying is None:
            raise NotImplementedError("No relation available for SelfReference")

    identifier = yaml_dict.get("identifier", 0)
    ref = ops.SelfReference(underlying, identifier=identifier)

    return ref.to_expr()


@register_from_yaml_handler("JoinReference")
def _join_reference_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    table_yaml = yaml_dict["parent"]
    return context.translate_from_yaml(table_yaml)


@register_from_yaml_handler("StructField")
def _structfield_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = tuple(context.translate_from_yaml(arg) for arg in yaml_dict["args"])
    return ops.StructField(*args).to_expr()


@translate_to_yaml.register(FrozenDict)
@translate_to_yaml.register(FrozenOrderedDict)
def _frozenordereddict_to_yaml(dct: dict, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(dct).__name__,
        }
        | {
            key: context.translate_to_yaml(value)
            if context is not None
            else translate_to_yaml(value, context)
            for key, value in dct.items()
        }
    )


@register_from_yaml_handler("FrozenOrderedDict")
def _frozenordereddict_from_yaml(
    yaml_dict: dict, context: TranslationContext
) -> FrozenOrderedDict:
    dct = FrozenOrderedDict(
        {
            key: context.translate_from_yaml(value)
            for key, value in toolz.dissoc(yaml_dict, "op").items()
        }
    )
    return dct


@register_from_yaml_handler("FrozenDict")
def _frozendict_from_yaml(yaml_dict: dict, context: TranslationContext) -> FrozenDict:
    dct = FrozenDict(
        {
            key: context.translate_from_yaml(value)
            for key, value in toolz.dissoc(yaml_dict, "op").items()
        }
    )
    return dct


@translate_to_yaml.register(Tag)
@convert_to_node_ref
def _tag_to_yaml(op: Tag, context: Any) -> dict:
    return freeze(
        {
            "op": "Tag",
            "parent": context.translate_to_yaml(op.parent),
            "metadata": context.translate_to_yaml(op.metadata),
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("Tag")
def _tag_from_yaml(yaml_dict: dict, context: Any) -> ibis.Expr:
    schema = context.get_schema(yaml_dict[RefEnum.schema_ref])

    # fixme: enable translation of nodes
    parent_expr = context.translate_from_yaml(yaml_dict["parent"])
    metadata = context.translate_from_yaml(yaml_dict["metadata"])
    op = Tag(
        schema=schema,
        parent=parent_expr.op(),
        metadata=metadata,
    )
    return op.to_expr()


@translate_to_yaml.register(ops.Argument)
def _array_filter_to_yaml(op: ops.Argument, context: Any) -> dict:
    return freeze(
        {
            "op": "Argument",
            "name": context.translate_to_yaml(op.name),
            "shape": context.translate_to_yaml(op.shape),
            "dtype": context.translate_to_yaml(op.dtype),
        }
    )


@register_from_yaml_handler("Argument")
def _array_filter_from_yaml(yaml_dict: dict, context: Any) -> ibis.Expr:
    name = context.translate_from_yaml(yaml_dict["name"])
    shape = context.translate_from_yaml(yaml_dict["shape"])
    dtype = context.translate_from_yaml(yaml_dict["dtype"])

    return ops.Argument(name, shape=shape, dtype=dtype).to_expr()


@translate_to_yaml.register(ds.DataShape)
def _columnar_to_yaml(op, context) -> dict:
    return freeze(
        {
            "op": op.__class__.__name__,
            "ndim": context.translate_to_yaml(op.ndim),
        }
    )


@register_from_yaml_handler("Columnar")
def _array_filter_from_yaml(yaml_dict: dict, context: Any) -> Any:
    return Columnar()


@translate_to_yaml.register(Schema)
@convert_to_schema_ref
def _schema_to_yaml(schema: Schema, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": schema.__class__.__name__,
            "value": freeze(toolz.valmap(context.translate_to_yaml, schema)),
        }
    )


@register_from_yaml_handler(Schema.__name__)
def _schema_from_yaml(yaml_dict: dict, context: TranslationContext) -> Schema:
    return Schema(toolz.valmap(context.translate_from_yaml, yaml_dict["value"]))


# sklearn estimator YAML translation
# Registration happens at module import time if sklearn is available
try:
    from sklearn.base import BaseEstimator

    @translate_to_yaml.register(BaseEstimator)
    def _sklearn_estimator_to_yaml(
        obj: BaseEstimator, context: TranslationContext
    ) -> dict:
        params = freeze(obj.get_params(deep=False))
        return freeze(
            {
                "op": "SklearnEstimator",
                "estimator_module": type(obj).__module__,
                "estimator_class": type(obj).__name__,
                "params": {k: context.translate_to_yaml(v) for k, v in params.items()},
            }
        )

except ImportError:
    pass


def _import_estimator_class(module: str, classname: str):
    """Import an sklearn estimator class from module and class name."""
    import importlib

    mod = importlib.import_module(module)
    return getattr(mod, classname)


@register_from_yaml_handler("SklearnEstimator")
def _sklearn_estimator_from_yaml(yaml_dict: dict, context: TranslationContext) -> Any:
    cls = _import_estimator_class(
        yaml_dict["estimator_module"], yaml_dict["estimator_class"]
    )
    params = {k: context.translate_from_yaml(v) for k, v in yaml_dict["params"].items()}
    return cls(**params)



import time

import pytest

import xorq.api as xo
from xorq.caching import (
    ParquetSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)
from xorq.common.utils.postgres_utils import (
    do_analyze,
    get_postgres_n_scans,
)


@pytest.mark.parametrize(
    "name",
    (
        "diamonds",
        "astronauts",
    ),
)
def test_source_caching(name, pg, parquet_dir):
    con = xo.connect()
    example = xo.deferred_read_parquet(parquet_dir / f"{name}.parquet", con)
    expr = example.cache(cache=SourceCache.from_kwargs(source=pg))
    assert not expr.ls.exists()
    actual = expr.execute()
    expected = example.execute()
    cached = pg.table(expr.ls.get_key()).execute()
    assert actual.equals(expected)
    assert actual.equals(cached)
    assert expr.ls.exists()
    pg.drop_table(expr.ls.get_key())
    assert not expr.ls.exists()


def test_postgres_cache_invalidation(pg, con):
    def modify_postgres_table(dt):
        (con, name) = (dt.source, dt.name)
        statement = f"""
        INSERT INTO "{name}"
        DEFAULT VALUES
        """
        con.raw_sql(statement)

    def assert_n_scans_changes(dt, n_scans_before):
        do_analyze(dt.source, dt.name)
        for _ in range(10):  # noqa: F402
            # give postgres some time to update its tables
            time.sleep(0.1)
            n_scans_after = get_postgres_n_scans(dt)
            if n_scans_before != n_scans_after:
                return n_scans_after
        else:
            raise

    (from_name, to_name) = ("batting", "batting_to_modify")
    if to_name in pg.tables:
        pg.drop_table(to_name)
    pg_t = pg.create_table(to_name, obj=pg.table(from_name).limit(1000))
    expr_cached = (
        pg_t.group_by("playerID")
        .size()
        .order_by("playerID")
        .cache(cache=SourceCache.from_kwargs(source=con))
    )
    dt = pg_t.op()
    (cache, uncached) = (expr_cached.ls.cache, expr_cached.ls.uncached_one)

    # assert initial state
    assert not cache.exists(uncached)
    n_scans_before = get_postgres_n_scans(dt)
    assert n_scans_before == 0

    # assert first execution state
    expr_cached.execute()
    n_scans_after = assert_n_scans_changes(dt, n_scans_before)
    # should we test that SourceCache.get is called?
    assert n_scans_after == 1
    assert cache.exists(uncached)

    # assert no change after re-execution of cached expr
    expr_cached.execute()
    assert n_scans_after == get_postgres_n_scans(dt)

    # assert cache invalidation happens
    modify_postgres_table(dt)
    expr_cached.execute()
    assert_n_scans_changes(dt, n_scans_after)


def test_postgres_snapshot(pg, con):
    def modify_postgres_table(dt):
        (con, name) = (dt.source, dt.name)
        statement = f"""
        INSERT INTO "{name}"
        DEFAULT VALUES
        """
        con.raw_sql(statement)

    def assert_n_scans_changes(dt, n_scans_before):
        do_analyze(dt.source, dt.name)
        for _ in range(10):  # noqa: F402
            # give postgres some time to update its tables
            time.sleep(0.1)
            n_scans_after = get_postgres_n_scans(dt)
            if n_scans_before != n_scans_after:
                return n_scans_after
        else:
            raise

    (from_name, to_name) = ("batting", "batting_to_modify")
    if to_name in pg.tables:
        pg.drop_table(to_name)
    pg_t = pg.create_table(to_name, obj=pg.table(from_name).limit(1000))
    cache = SourceSnapshotCache.from_kwargs(source=con)
    expr_cached = (
        pg_t.group_by("playerID").size().order_by("playerID").cache(cache=cache)
    )
    dt = pg_t.op()
    (cache, uncached) = (expr_cached.ls.cache, expr_cached.ls.uncached_one)

    # assert initial state
    assert not cache.exists(uncached)
    n_scans_before = get_postgres_n_scans(dt)
    assert n_scans_before == 0

    # assert first execution state
    executed0 = expr_cached.execute()
    n_scans_after = assert_n_scans_changes(dt, n_scans_before)
    # should we test that SourceCache.get is called?
    assert n_scans_after == 1
    assert cache.exists(uncached)

    # assert no change after re-execution of cached expr
    executed1 = expr_cached.execute()
    assert n_scans_after == get_postgres_n_scans(dt)
    assert executed0.equals(executed1)

    # assert NO cache invalidation
    modify_postgres_table(dt)
    executed2 = expr_cached.execute()
    assert executed0.equals(executed2)
    with pytest.raises(Exception):  # noqa: B017
        assert_n_scans_changes(dt, n_scans_after)

    executed3 = expr_cached.ls.uncached.execute()
    assert not executed0.equals(executed3)


def test_postgres_parquet_snapshot(pg, tmp_path):
    def modify_postgres_table(dt):
        (con, name) = (dt.source, dt.name)
        statement = f"""
        INSERT INTO "{name}"
        DEFAULT VALUES
        """
        con.raw_sql(statement)

    def assert_n_scans_changes(dt, n_scans_before):
        do_analyze(dt.source, dt.name)
        for _i in range(10):  # noqa: F402
            # give postgres some time to update its tables
            time.sleep(0.1)
            n_scans_after = get_postgres_n_scans(dt)
            if n_scans_before != n_scans_after:
                return n_scans_after
        else:
            raise

    (from_name, to_name) = ("batting", "batting_to_modify")
    if to_name in pg.tables:
        pg.drop_table(to_name)
    pg_t = pg.create_table(to_name, obj=pg.table(from_name).limit(1000))
    cache = ParquetSnapshotCache.from_kwargs(
        relative_path=tmp_path.joinpath("parquet-snapshot-storage")
    )
    expr = pg_t.group_by("playerID").size().order_by("playerID")
    expr_cached = expr.cache(cache=cache)
    dt = pg_t.op()
    (cache, uncached) = (expr_cached.ls.cache, expr_cached.ls.uncached_one)

    # assert initial state
    assert not cache.exists(uncached)
    n_scans_before = get_postgres_n_scans(dt)
    assert n_scans_before == 0

    # assert first execution state
    executed0 = expr_cached.execute()
    n_scans_after = assert_n_scans_changes(dt, n_scans_before)
    # should we test that SourceCache.get is called?
    assert n_scans_after == 1
    assert cache.exists(uncached)
    assert cache.exists(expr)

    # assert no change after re-execution of cached expr
    executed1 = expr_cached.execute()
    assert n_scans_after == get_postgres_n_scans(dt)
    assert executed0.equals(executed1)

    # assert NO cache invalidation
    modify_postgres_table(dt)
    executed2 = expr_cached.execute()
    assert executed0.equals(executed2)
    with pytest.raises(Exception):  # noqa: B017
        assert_n_scans_changes(dt, n_scans_after)

    executed3 = expr_cached.ls.uncached.execute()
    assert not executed0.equals(executed3)



import operator

import toolz

import xorq.api as xo
from xorq.api import make_pandas_udf


field_name = "price_per_carat"
schema = xo.schema({"sum_price": "float64", "sum_carat": "float64"})
return_type = xo.dtype("float64")


@make_pandas_udf(
    schema=schema,
    return_type=return_type,
    name="my_udf",
)
def my_udf(df):
    return df["sum_carat"].div(df["sum_price"])


def do_agg(expr):
    return (
        expr.group_by("cut")
        .agg(
            xo._.price.sum().name("sum_price"),
            xo._.carat.sum().name("sum_carat"),
            xo._.color.nunique().name("nunique_color"),
        )
        .cast({"sum_price": "float64"})
    )


my_udf_on_expr = toolz.compose(
    operator.methodcaller("name", field_name), my_udf.on_expr
)

con = xo.connect()

diamonds = xo.examples.diamonds.fetch(backend=con)

input_expr = diamonds.pipe(do_agg)
process_df = operator.methodcaller("assign", **{field_name: my_udf.fn})
maybe_schema_in = input_expr.schema()
maybe_schema_out = xo.schema(input_expr.schema() | {field_name: return_type})
command = "diamonds_exchange_command"

expr = xo.expr.relations.flight_udxf(
    input_expr,
    process_df=process_df,
    maybe_schema_in=maybe_schema_in,
    maybe_schema_out=maybe_schema_out,
    con=con,
    make_udxf_kwargs={"name": my_udf.__name__, "command": command},
).order_by("cut")



import pandas as pd
import pyarrow as pa

import xorq.api as xo
import xorq.flight.exchanger as E
from xorq.flight import FlightServer
from xorq.flight.action import (
    DropTableAction,
    GetExchangeAction,
    GetSchemaQueryAction,
    ListTablesAction,
)


def test_list_tables_kwargs():
    with FlightServer(
        verify_client=False,
        make_connection=xo.duckdb.connect,
    ) as main:
        # GIVEN
        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()
        main.con.create_table("users", data)
        expected = ["users"]

        # WHEN
        actual = main.client.do_action_one(ListTablesAction.name)

        # THEN
        assert actual == expected


def test_drop_table():
    with FlightServer(
        verify_client=False,
        make_connection=xo.duckdb.connect,
    ) as main:
        # GIVEN
        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()
        main.con.register(data, table_name="users")
        assert main.client.do_action_one(ListTablesAction.name)

        # WHEN
        main.client.do_action_one(DropTableAction.name, action_body={"name": "users"})

        # THEN
        assert not main.client.do_action_one(ListTablesAction.name)


def test_get_schema_query():
    with FlightServer(
        verify_client=False,
        make_connection=xo.duckdb.connect,
    ) as main:
        # GIVEN
        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()
        main.con.register(data, table_name="users")
        expected = xo.Schema(
            {
                "id": pa.int64(),
                "name": pa.string(),
            }
        )
        # WHEN
        actual = main.client.do_action_one(
            GetSchemaQueryAction.name, action_body="SELECT * FROM users;"
        )
        # THEN
        assert actual == expected


def test_get_exchange():
    def dummy(df: pd.DataFrame):
        return pd.DataFrame({"row_count": [42]})

    schema_in = xo.schema({"dummy": "int64"})
    dummy_udxf = E.make_udxf(
        dummy,
        schema_in,
        xo.schema({"row_count": "int64"}),
    )
    with FlightServer(exchangers=[dummy_udxf]) as flight_server:
        client = flight_server.client
        exchange = client.do_action_one(
            GetExchangeAction.name, action_body=dummy_udxf.command
        )
        assert exchange is not None



import operator

import pytest

from xorq.vendor import ibis


try:
    from duckdb import InvalidInputException as DuckDBInvalidInputException
except ImportError:
    DuckDBInvalidInputException = None

pd = pytest.importorskip("pandas")
tm = pytest.importorskip("pandas.testing")


@pytest.fixture(scope="module")
def time_keyed_df1():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=6)
            ),
            "key": [1, 2, 3, 1, 2, 3],
            "value": [1.2, 1.4, 2.0, 4.0, 8.0, 16.0],
        }
    )


@pytest.fixture(scope="module")
def time_keyed_df2():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", freq="3D", periods=3)
            ),
            "key": [1, 2, 3],
            "other_value": [1.1, 1.2, 2.2],
        }
    )


@pytest.fixture(scope="module")
def time_keyed_left(time_keyed_df1):
    return ibis.memtable(time_keyed_df1)


@pytest.fixture(scope="module")
def time_keyed_right(time_keyed_df2):
    return ibis.memtable(time_keyed_df2)


@pytest.mark.parametrize(
    ("direction", "op"), [("backward", operator.ge), ("forward", operator.le)]
)
@pytest.mark.parametrize("right_column_key", [None, "on_right_time"])
@pytest.mark.xfail_version(
    duckdb=["duckdb>=0.10.2,<1.1.1"], raises=DuckDBInvalidInputException
)
def test_keyed_asof_join_with_tolerance(
    duckdb_con,
    time_keyed_left,
    time_keyed_right,
    time_keyed_df1,
    time_keyed_df2,
    direction,
    op,
    right_column_key,
):
    left_key = right_key = "time"
    if right_column_key:
        right_key = right_column_key
        time_keyed_right = time_keyed_right.rename({right_column_key: "time"})
        assert left_key != right_key

    on = op(time_keyed_left[left_key], time_keyed_right[right_key])
    expr = time_keyed_left.asof_join(
        time_keyed_right, on, "key", tolerance=ibis.interval(days=2)
    )

    result = duckdb_con.execute(expr)
    expected = pd.merge_asof(
        time_keyed_df1,
        time_keyed_df2,
        on="time",
        by="key",
        tolerance=pd.Timedelta("2D"),
        direction=direction,
    )

    result = result.sort_values(["key", "time"]).reset_index(drop=True)
    expected = expected.sort_values(["key", "time"]).reset_index(drop=True)

    tm.assert_frame_equal(
        # drop `time` from comparison to avoid issues with different time resolution
        result[expected.columns].drop(["time"], axis=1),
        expected.drop(["time"], axis=1),
    )

    # check that time is equal in value, if not dtype
    tm.assert_series_equal(result["time"], expected["time"], check_dtype=False)



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import pandas as pd
import pyarrow as pa

from xorq.vendor.ibis.formats.pandas import PandasData
from xorq.vendor.ibis.formats.pyarrow import PyArrowData


if TYPE_CHECKING:
    import xorq.vendor.ibis.expr.datatypes as dt
    from xorq.vendor.ibis.expr.schema import Schema


class JSONScalar(pa.ExtensionScalar):
    def as_py(self):
        value = self.value
        if value is None:
            return value
        else:
            return json.loads(value.as_py())


class JSONArray(pa.ExtensionArray):
    pass


class JSONType(pa.ExtensionType):
    def __init__(self):
        super().__init__(pa.string(), "xorq.json")

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return cls()

    def __arrow_ext_class__(self):
        return JSONArray

    def __arrow_ext_scalar_class__(self):
        return JSONScalar


PYARROW_JSON_TYPE = JSONType()
pa.register_extension_type(PYARROW_JSON_TYPE)


class SnowflakePandasData(PandasData):
    @classmethod
    def convert_Timestamp_element(cls, dtype):
        return pd.Timestamp.fromisoformat

    @classmethod
    def convert_Date_element(cls, dtype):
        return pd.Timestamp.fromisoformat

    @classmethod
    def convert_Time_element(cls, dtype):
        return pd.Timestamp.fromisoformat

    @classmethod
    def convert_JSON(cls, s, dtype, pandas_type):
        converter = cls.convert_JSON_element(dtype)
        return s.map(converter, na_action="ignore").astype("object")

    @classmethod
    def convert_Array(cls, s, dtype, pandas_type):
        raw_json_objects = cls.convert_JSON(s, dtype, pandas_type)
        return super().convert_Array(raw_json_objects, dtype, pandas_type)

    @classmethod
    def convert_Map(cls, s, dtype, pandas_type):
        raw_json_objects = cls.convert_JSON(s, dtype, pandas_type)
        return super().convert_Map(raw_json_objects, dtype, pandas_type)

    @classmethod
    def convert_Struct(cls, s, dtype, pandas_type):
        raw_json_objects = cls.convert_JSON(s, dtype, pandas_type)
        return super().convert_Struct(raw_json_objects, dtype, pandas_type)


class SnowflakePyArrowData(PyArrowData):
    @classmethod
    def convert_table(cls, table: pa.Table, schema: Schema) -> pa.Table:
        columns = [cls.convert_column(table[name], typ) for name, typ in schema.items()]
        return pa.Table.from_arrays(
            columns,
            schema=pa.schema(
                [
                    pa.field(name, array.type, nullable=ibis_dtype.nullable)
                    for array, (name, ibis_dtype) in zip(columns, schema.items())
                ]
            ),
        )

    @classmethod
    def convert_column(cls, column: pa.Array, dtype: dt.DataType) -> pa.Array:
        if dtype.is_json() or dtype.is_array() or dtype.is_map() or dtype.is_struct():
            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()

            return pa.ExtensionArray.from_storage(PYARROW_JSON_TYPE, column)
        return super().convert_column(column, dtype)

    @classmethod
    def convert_scalar(cls, scalar: pa.Scalar, dtype: dt.DataType) -> pa.Scalar:
        if dtype.is_json() or dtype.is_array() or dtype.is_map() or dtype.is_struct():
            return pa.ExtensionScalar.from_storage(PYARROW_JSON_TYPE, scalar)
        return super().convert_scalar(scalar, dtype)



import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.operations as ops
from xorq.ibis_yaml.common import (
    RefEnum,
    RegistryEnum,
)
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_scalar_subquery(compiler, t):
    expr = ops.ScalarSubquery(t.c.mean().as_table()).to_expr()
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]

    assert expression["op"] == "ScalarSubquery"
    assert expression["args"][0]["op"] == "Aggregate"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_exists_subquery(compiler):
    t1 = ibis.table({"a": "int", "b": "string"}, name="t1")
    t2 = ibis.table({"a": "int", "c": "float"}, name="t2")

    filtered = t2.filter(t2.a == t1.a)
    expr = ops.ExistsSubquery(filtered).to_expr()
    yaml_dict = compiler.to_yaml(expr)

    expression = yaml_dict["expression"]

    assert expression["op"] == "ExistsSubquery"
    node_ref = expression["rel"][RefEnum.node_ref]
    assert yaml_dict["definitions"][RegistryEnum.nodes][node_ref]["op"] == "Filter"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_in_subquery(compiler):
    t1 = ibis.table({"a": "int", "b": "string"}, name="t1")
    t2 = ibis.table({"a": "int", "c": "float"}, name="t2")

    expr = ops.InSubquery(t1.select("a"), t2.a).to_expr()
    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)

    assert expression["op"] == "InSubquery"
    assert dtype_yaml["type"] == "Boolean"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)



import ast
import pathlib
from typing import Any, Optional, Union

from xorq.backends.xorq import connect
from xorq.common.utils.env_utils import (
    EnvConfigable,
    env_templates_dir,
)
from xorq.vendor import ibis
from xorq.vendor.ibis.config import Config
from xorq.vendor.ibis.config import Options as IbisOptions


env_config = EnvConfigable.subclass_from_env_file(
    env_templates_dir.joinpath(".env.xorq.template")
).from_env()


class Cache(Config):
    """xorq cache configuration options

    Attributes
    ----------

    default_relative_path : str

    """

    default_relative_path: Union[str, pathlib.Path] = pathlib.Path(
        env_config.XORQ_DEFAULT_RELATIVE_PATH
    )
    key_prefix: str = env_config.XORQ_CACHE_KEY_PREFIX


class Interactive(Config):
    """Options controlling the interactive repr."""

    @property
    def max_rows(self) -> int:
        return ibis.options.repr.interactive.max_rows

    @max_rows.setter
    def max_rows(self, value: int):
        ibis.options.repr.interactive.max_rows = value

    @property
    def max_columns(self) -> Optional[int]:
        return ibis.options.repr.interactive.max_columns

    @max_columns.setter
    def max_columns(self, value: Optional[int]):
        ibis.options.repr.interactive.max_columns = value

    @property
    def max_length(self) -> int:
        return ibis.options.repr.interactive.max_length

    @max_length.setter
    def max_length(self, value: int):
        ibis.options.repr.interactive.max_length = value

    @property
    def max_string(self) -> int:
        return ibis.options.repr.interactive.max_string

    @max_string.setter
    def max_string(self, value: int):
        ibis.options.repr.interactive.max_string = value

    @property
    def max_depth(self) -> int:
        return ibis.options.repr.interactive.max_depth

    @max_depth.setter
    def max_depth(self, value: int):
        ibis.options.repr.interactive.max_depth = value

    @property
    def show_types(self) -> bool:
        return ibis.options.repr.interactive.show_types

    @show_types.setter
    def show_types(self, value: bool):
        ibis.options.repr.interactive.show_types = value


class Repr(Config):
    """Expression printing options.

    Attributes
    ----------
    interactive : Interactive
        Options controlling the interactive repr.
    """

    interactive: Interactive = Interactive()


class SQL(Config):
    """SQL-related options.

    Attributes
    ----------
    dialect : str
        Dialect to use for printing SQL when the backend cannot be determined.

    """

    dialect: str = "datafusion"


class Pins(Config):
    """SQL-related options.

    Attributes
    ----------
    dialect : str
        Dialect to use for printing SQL when the backend cannot be determined.

    """

    protocol: str = "gcs"
    path: str = "letsql-pins"
    storage_options: dict[str, Any] = {
        "cache_timeout": 0,
        "token": "anon",
    }

    def get_board(self, **kwargs):
        import pins

        _kwargs = {
            **{
                "protocol": self.protocol,
                "path": self.path,
                "storage_options": self.storage_options,
            },
            **kwargs,
        }
        return pins.board(**_kwargs)

    def get_path(self, name, board=None, **kwargs):
        import tempfile

        from filelock import FileLock

        board = board or self.get_board()

        # Cross-process lock keyed by pin name to prevent concurrent
        # downloads from corrupting cached files under pytest-xdist.
        lock_path = pathlib.Path(tempfile.gettempdir()) / f"xorq-pin-{name}.lock"
        lock = FileLock(lock_path, timeout=120)

        with lock:
            (path,) = board.pin_download(name, **kwargs)

        return path


class Options(IbisOptions):
    """xorq configuration options

    Attributes
    ----------
    cache : Cache
        Options controlling caching.
    backend : Optional[xorq.backends.xorq.Backend]
        The backend to use for execution.
    repr : Repr
        Options controlling expression printing.
    """

    cache: Cache = Cache()
    backend: Optional[Any] = None
    repr: Repr = Repr()
    sql: SQL = SQL()
    pins: Pins = Pins()
    debug: bool = bool(ast.literal_eval(env_config.XORQ_DEBUG or 0))

    @property
    def interactive(self) -> bool:
        """Show the first few rows of computing an expression when in a repl."""
        return ibis.options.interactive

    @interactive.setter
    def interactive(self, value: bool):
        ibis.options.interactive = value


options = Options()


def _backend_init():
    if (backend := options.backend) is not None:
        return backend

    options.backend = con = connect()
    return con



import itertools
import traceback

import pyarrow as pa
import pyarrow.compute as pc
import toolz
from opentelemetry import trace

from xorq.common.utils.otel_utils import (
    tracer,
)


@toolz.curry
def excepts_print_exc(func, exc=Exception, handler=toolz.functoolz.return_none):
    _handler = toolz.compose(handler, toolz.curried.do(traceback.print_exception))
    return toolz.excepts(exc, func, _handler)


@tracer.start_as_current_span("otel_instrument_reader")
def otel_instrument_reader(reader):
    span = trace.get_current_span()
    ctx = span.get_span_context()
    span_link = trace.Link(ctx)
    event_ids = {"reader_id": id(reader), "trace_id": f"{ctx.trace_id:x}"}
    span.add_event(
        "span.metrics.rbr.make_reader",
        event_ids,
    )

    def record_metrics(**kwargs):
        # we can't directly instrument a generator
        # https://github.com/open-telemetry/opentelemetry-python/issues/2606#issuecomment-2405109375
        with tracer.start_as_current_span(
            "otel_instrument_reader.metrics_recorder", links=[span_link]
        ) as span:
            span.add_event(
                "span.metrics.rbr.record_reader",
                kwargs | event_ids,
            )

    def instrument_reader(reader):
        n_batches = 0
        sum_buffer_size = 0
        for batch in reader:
            n_batches += 1
            sum_buffer_size += batch.get_total_buffer_size()
            yield batch
        record_metrics(n_batches=n_batches, sum_buffer_size=sum_buffer_size)

    return pa.RecordBatchReader.from_batches(reader.schema, instrument_reader(reader))


def copy_rbr_batches(reader):
    manager = pa.default_cpu_memory_manager()
    gen = (batch.copy_to(manager) for batch in reader)
    return pa.RecordBatchReader.from_batches(reader.schema, gen)


def make_filtered_reader(reader):
    gen = (chunk.data for chunk in reader if chunk.data)
    return pa.RecordBatchReader.from_batches(reader.schema, gen)


def instrument_reader(reader, prefix=""):
    from xorq.common.utils.logging_utils import get_print_logger

    logger = get_print_logger()

    def gen(reader):
        logger.info(f"{prefix:10s}first batch yielded")
        yield next(reader)
        yield from reader
        logger.info(f"{prefix:10s}last batch yielded")

    return pa.RecordBatchReader.from_batches(reader.schema, gen(reader))


@excepts_print_exc
def streaming_split_exchange(
    split_key, f, context, reader, writer, options=None, **kwargs
):
    started = False
    g = excepts_print_exc(f)
    for split_reader in ReaderSplitter(
        make_filtered_reader(reader),
        split_key,
    ):
        batch = g(split_reader)
        if not started:
            writer.begin(batch.schema, options=options)
            started = True
        writer.write_batch(batch)


class ReaderSplitter:
    """Yield multiple record batch readers from a single record batch reader by `split_key`"""

    def __init__(self, rbr, split_key):
        self.rbr = rbr
        self.split_key = split_key
        self._curr_batch = toolz.excepts(StopIteration, next)(self.rbr)

    @property
    def curr_batch(self):
        if not self._curr_batch or not len(self._curr_batch):
            return
        return self._curr_batch

    @property
    def curr_key(self):
        if not self.curr_batch:
            raise ValueError("no current batch available; reader may be exhausted")
        (curr_key,) = self.curr_batch[self.split_key].slice(length=1).tolist()
        return curr_key

    def split_batch(self, batch, curr_key):
        condition = pc.field(self.split_key) == curr_key
        try:
            good_batch = batch.filter(condition)
        except IndexError:
            return (None, batch)
        if len(good_batch) == len(batch):
            return (batch, None)
        else:
            bad_batch = batch.filter(~condition)
            return (good_batch, bad_batch)

    @property
    def next_split(self):
        batches = ()
        curr_key = self.curr_key
        (curr_batch, self._curr_batch) = (self.curr_batch, None)
        for batch in itertools.chain([curr_batch], self.rbr):
            good_batch, bad_batch = self.split_batch(batch, curr_key)
            if good_batch:
                batches += (good_batch,)
            if bad_batch:
                self._curr_batch = bad_batch
                return pa.RecordBatchReader.from_batches(
                    curr_batch.schema,
                    batches,
                )
        # fall through condition
        return pa.RecordBatchReader.from_batches(
            curr_batch.schema,
            batches,
        )

    def __next__(self):
        return self.next_split

    def __iter__(self):
        while self.curr_batch:
            yield self.next_split

    @classmethod
    def gen_splits(cls, expr, split_key):
        rbr = expr.order_by(split_key).to_pyarrow_batches()
        yield from cls(rbr, split_key)



from __future__ import annotations

from types import FunctionType, LambdaType  # noqa: TCH003
from typing import Union

from public import public

import xorq.vendor.ibis.expr.datashape as ds
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.operations.analytic import Analytic
from xorq.vendor.ibis.expr.operations.core import Column, Value
from xorq.vendor.ibis.expr.operations.reductions import Reduction


class VectorizedUDF(Value):
    func: Union[FunctionType, LambdaType]
    func_args: VarTuple[Column]
    # TODO(kszucs): should rename these arguments to input_dtypes and return_dtype
    input_type: VarTuple[dt.DataType]
    return_type: dt.DataType

    @property
    def dtype(self):
        return self.return_type


@public
class ElementWiseVectorizedUDF(VectorizedUDF):
    """Node for element wise UDF."""

    shape = ds.columnar


@public
class ReductionVectorizedUDF(VectorizedUDF, Reduction):
    """Node for reduction UDF."""

    shape = ds.scalar


# TODO(kszucs): revisit
@public
class AnalyticVectorizedUDF(VectorizedUDF, Analytic):
    """Node for analytics UDF."""

    shape = ds.columnar
